# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pytest

import re

import numpy as np

from hatchet import GraphFrame
from hatchet.node import traversal_order
from hatchet.query import (
    QueryMatcher,
    InvalidQueryFilter,
    InvalidQueryPath,
    AbstractQuery,
    NaryQuery,
    AndQuery,
    OrQuery,
    XorQuery,
    IntersectionQuery,
    UnionQuery,
    SymDifferenceQuery,
    CypherQuery,
)


def test_construct_high_level_api():
    mock_node_mpi = {"name": "MPI_Bcast"}
    mock_node_ibv = {"name": "ibv_reg_mr"}
    mock_node_time_true = {"time (inc)": 0.1}
    mock_node_time_false = {"time (inc)": 0.001}
    path1 = [{"name": "MPI_[_a-zA-Z]*"}, "*", {"name": "ibv[_a-zA-Z]*"}]
    path2 = [{"name": "MPI_[_a-zA-Z]*"}, 2, {"name": "ibv[_a-zA-Z]*"}]
    path3 = [
        {"name": "MPI_[_a-zA-Z]*"},
        ("+", {"time (inc)": ">= 0.1"}),
        {"name": "ibv[_a-zA-Z]*"},
    ]
    path4 = [
        {"name": "MPI_[_a-zA-Z]*"},
        (3, {"time (inc)": 0.1}),
        {"name": "ibv[_a-zA-Z]*"},
    ]
    query1 = QueryMatcher(path1)
    query2 = QueryMatcher(path2)
    query3 = QueryMatcher(path3)
    query4 = QueryMatcher(path4)

    assert query1.query_pattern[0][0] == "."
    assert query1.query_pattern[0][1](mock_node_mpi)
    assert not query1.query_pattern[0][1](mock_node_ibv)
    assert not query1.query_pattern[0][1](mock_node_time_true)
    assert query1.query_pattern[1][0] == "*"
    assert query1.query_pattern[1][1](mock_node_mpi)
    assert query1.query_pattern[1][1](mock_node_ibv)
    assert query1.query_pattern[1][1](mock_node_time_true)
    assert query1.query_pattern[1][1](mock_node_time_false)
    assert query1.query_pattern[2][0] == "."

    assert query2.query_pattern[0][0] == "."
    assert query2.query_pattern[1][0] == "."
    assert query2.query_pattern[1][1](mock_node_mpi)
    assert query2.query_pattern[1][1](mock_node_ibv)
    assert query2.query_pattern[1][1](mock_node_time_true)
    assert query2.query_pattern[1][1](mock_node_time_false)
    assert query2.query_pattern[2][0] == "."
    assert query2.query_pattern[2][1](mock_node_mpi)
    assert query2.query_pattern[2][1](mock_node_ibv)
    assert query2.query_pattern[2][1](mock_node_time_true)
    assert query2.query_pattern[2][1](mock_node_time_false)
    assert query2.query_pattern[3][0] == "."

    assert query3.query_pattern[0][0] == "."
    assert query3.query_pattern[1][0] == "+"
    assert not query3.query_pattern[1][1](mock_node_mpi)
    assert not query3.query_pattern[1][1](mock_node_ibv)
    assert query3.query_pattern[1][1](mock_node_time_true)
    assert not query3.query_pattern[1][1](mock_node_time_false)
    assert query3.query_pattern[2][0] == "."

    assert query4.query_pattern[0][0] == "."
    assert query4.query_pattern[1][0] == "."
    assert not query4.query_pattern[1][1](mock_node_mpi)
    assert not query4.query_pattern[1][1](mock_node_ibv)
    assert query4.query_pattern[1][1](mock_node_time_true)
    assert not query4.query_pattern[1][1](mock_node_time_false)
    assert query4.query_pattern[2][0] == "."
    assert not query4.query_pattern[2][1](mock_node_mpi)
    assert not query4.query_pattern[2][1](mock_node_ibv)
    assert query4.query_pattern[2][1](mock_node_time_true)
    assert not query4.query_pattern[2][1](mock_node_time_false)
    assert query4.query_pattern[3][0] == "."
    assert not query4.query_pattern[3][1](mock_node_mpi)
    assert not query4.query_pattern[3][1](mock_node_ibv)
    assert query4.query_pattern[3][1](mock_node_time_true)
    assert not query4.query_pattern[3][1](mock_node_time_false)
    assert query4.query_pattern[4][0] == "."

    invalid_path = [
        {"name": "MPI_[_a-zA-Z]*"},
        ({"bad": "wildcard"}, {"time (inc)": 0.1}),
        {"name": "ibv[_a-zA-Z]*"},
    ]
    with pytest.raises(InvalidQueryPath):
        _ = QueryMatcher(invalid_path)

    invalid_path = [["name", "MPI_[_a-zA-Z]*"], "*", {"name": "ibv[_a-zA-Z]*"}]
    with pytest.raises(InvalidQueryPath):
        _ = QueryMatcher(invalid_path)


def test_construct_low_level_api():
    mock_node_mpi = {"name": "MPI_Bcast"}
    mock_node_ibv = {"name": "ibv_reg_mr"}
    mock_node_time_true = {"time (inc)": 0.1}
    mock_node_time_false = {"time (inc)": 0.001}

    def mpi_filter(df_row):
        if "name" not in df_row:
            return False
        if re.match(r"MPI_[_a-zA-Z]*\Z", df_row["name"]) is not None:
            return True
        return False

    def ibv_filter(df_row):
        if "name" not in df_row:
            return False
        if re.match(r"ibv[_a-zA-Z]*\Z", df_row["name"]) is not None:
            return True
        return False

    def time_ge_filter(df_row):
        if "time (inc)" not in df_row:
            return False
        return df_row["time (inc)"] >= 0.1

    def time_eq_filter(df_row):
        if "time (inc)" not in df_row:
            return False
        return df_row["time (inc)"] == 0.1

    query = QueryMatcher()

    query.match(filter_func=mpi_filter).rel("*").rel(filter_func=ibv_filter)
    assert query.query_pattern[0][0] == "."
    assert query.query_pattern[0][1](mock_node_mpi)
    assert not query.query_pattern[0][1](mock_node_ibv)
    assert not query.query_pattern[0][1](mock_node_time_true)
    assert query.query_pattern[1][0] == "*"
    assert query.query_pattern[1][1](mock_node_mpi)
    assert query.query_pattern[1][1](mock_node_ibv)
    assert query.query_pattern[1][1](mock_node_time_true)
    assert query.query_pattern[1][1](mock_node_time_false)
    assert query.query_pattern[2][0] == "."

    query.match(filter_func=mpi_filter).rel(2).rel(filter_func=ibv_filter)
    assert query.query_pattern[0][0] == "."
    assert query.query_pattern[1][0] == "."
    assert query.query_pattern[1][1](mock_node_mpi)
    assert query.query_pattern[1][1](mock_node_ibv)
    assert query.query_pattern[1][1](mock_node_time_true)
    assert query.query_pattern[1][1](mock_node_time_false)
    assert query.query_pattern[2][0] == "."
    assert query.query_pattern[2][1](mock_node_mpi)
    assert query.query_pattern[2][1](mock_node_ibv)
    assert query.query_pattern[2][1](mock_node_time_true)
    assert query.query_pattern[2][1](mock_node_time_false)
    assert query.query_pattern[3][0] == "."

    query.match(filter_func=mpi_filter).rel("+", time_ge_filter).rel(
        filter_func=ibv_filter
    )
    assert query.query_pattern[0][0] == "."
    assert query.query_pattern[1][0] == "+"
    assert not query.query_pattern[1][1](mock_node_mpi)
    assert not query.query_pattern[1][1](mock_node_ibv)
    assert query.query_pattern[1][1](mock_node_time_true)
    assert not query.query_pattern[1][1](mock_node_time_false)
    assert query.query_pattern[2][0] == "."

    query.match(filter_func=mpi_filter).rel(3, time_eq_filter).rel(
        filter_func=ibv_filter
    )
    assert query.query_pattern[0][0] == "."
    assert query.query_pattern[1][0] == "."
    assert not query.query_pattern[1][1](mock_node_mpi)
    assert not query.query_pattern[1][1](mock_node_ibv)
    assert query.query_pattern[1][1](mock_node_time_true)
    assert not query.query_pattern[1][1](mock_node_time_false)
    assert query.query_pattern[2][0] == "."
    assert not query.query_pattern[2][1](mock_node_mpi)
    assert not query.query_pattern[2][1](mock_node_ibv)
    assert query.query_pattern[2][1](mock_node_time_true)
    assert not query.query_pattern[2][1](mock_node_time_false)
    assert query.query_pattern[3][0] == "."
    assert not query.query_pattern[3][1](mock_node_mpi)
    assert not query.query_pattern[3][1](mock_node_ibv)
    assert query.query_pattern[3][1](mock_node_time_true)
    assert not query.query_pattern[3][1](mock_node_time_false)
    assert query.query_pattern[4][0] == "."


def test_node_caching(mock_graph_literal):
    path = [{"name": "fr[a-z]+"}, ("+", {"time (inc)": ">= 25.0"}), {"name": "baz"}]
    gf = GraphFrame.from_literal(mock_graph_literal)
    node = gf.graph.roots[0].children[2].children[0]

    query = QueryMatcher(path)
    query._cache_node(gf, node)

    assert 0 in query.search_cache[node._hatchet_nid]
    assert 1 in query.search_cache[node._hatchet_nid]
    assert 2 not in query.search_cache[node._hatchet_nid]


def test_match_0_or_more_wildcard(mock_graph_literal):
    path = [
        {"name": "qux"},
        ("*", {"time (inc)": "> 10"}),
        {"name": "gr[a-z]+", "time (inc)": "<= 10"},
    ]
    gf = GraphFrame.from_literal(mock_graph_literal)
    node = gf.graph.roots[0].children[1]
    none_node = gf.graph.roots[0].children[2].children[0].children[1].children[0]

    correct_paths = [
        [
            node.children[0],
            node.children[0].children[0],
            node.children[0].children[0].children[0],
        ],
        [node.children[0], node.children[0].children[0]],
    ]

    query = QueryMatcher(path)
    matched_paths = []
    for child in sorted(node.children, key=traversal_order):
        match = query._match_0_or_more(gf, child, 1)
        if match is not None:
            matched_paths.extend(match)

    assert sorted(matched_paths, key=len) == sorted(correct_paths, key=len)
    assert query._match_0_or_more(gf, none_node, 1) is None


def test_match_1_or_more_wildcard(mock_graph_literal):
    path = [
        {"name": "qux"},
        ("+", {"time (inc)": "> 10"}),
        {"name": "gr[a-z]+", "time (inc)": "<= 10"},
    ]
    gf = GraphFrame.from_literal(mock_graph_literal)
    node = gf.graph.roots[0].children[1]
    none_node = gf.graph.roots[0].children[2].children[0].children[1].children[0]

    correct_paths = [
        [
            node.children[0],
            node.children[0].children[0],
            node.children[0].children[0].children[0],
        ],
        [node.children[0], node.children[0].children[0]],
    ]

    query = QueryMatcher(path)
    matched_paths = []
    for child in sorted(node.children, key=traversal_order):
        match = query._match_1_or_more(gf, child, 1)
        if match is not None:
            matched_paths.extend(match)

    assert matched_paths == correct_paths
    assert query._match_1_or_more(gf, none_node, 1) is None

    zero_match_path = [
        {"name": "qux"},
        ("+", {"time (inc)": "> 50"}),
        {"name": "gr[a-z]+", "time (inc)": "<= 10"},
    ]
    zero_match_node = gf.graph.roots[0].children[0]
    query = QueryMatcher(zero_match_path)
    assert query._match_1_or_more(gf, zero_match_node, 1) is None


def test_match_1(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    path = [
        {"name": "qux"},
        ("*", {"time (inc)": "> 10"}),
        {"name": "gr[a-z]+", "time (inc)": "<= 10.0"},
    ]
    query = QueryMatcher(path)

    assert query._match_1(gf, gf.graph.roots[0].children[0], 2) == [
        [gf.graph.roots[0].children[0].children[1]]
    ]
    assert query._match_1(gf, gf.graph.roots[0], 2) is None


def test_match(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    root = gf.graph.roots[0].children[2]

    path0 = [
        {"name": "waldo"},
        "+",
        {"time (inc)": ">= 20.0"},
        "+",
        {"time (inc)": 5.0, "time": 5.0},
    ]
    match0 = [
        [
            root,
            root.children[0],
            root.children[0].children[1],
            root.children[0].children[1].children[0],
            root.children[0].children[1].children[0].children[0],
        ]
    ]
    query0 = QueryMatcher(path0)
    assert query0._match_pattern(gf, root, 0) == match0

    path1 = [
        {"name": "waldo"},
        ("+", {}),
        {"time (inc)": ">= 20.0"},
        "+",
        {"time (inc)": 7.5, "time": 7.5},
    ]
    query1 = QueryMatcher(path1)
    assert query1._match_pattern(gf, root, 0) is None


def test_apply(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    path = [
        {"time (inc)": ">= 30.0"},
        (2, {"name": "[^b][a-z]+"}),
        ("*", {"name": "[^b][a-z]+"}),
        {"name": "gr[a-z]+"},
    ]
    root = gf.graph.roots[0]
    match = [
        root,
        root.children[1],
        root.children[1].children[0],
        root.children[1].children[0].children[0],
        root.children[1].children[0].children[0].children[1],
        # Old-style return value of apply
        # [
        #     root,
        #     root.children[1],
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[1],
        # ],
        # [
        #     root.children[1],
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[1],
        # ],
    ]
    query = QueryMatcher(path)

    assert sorted(query.apply(gf)) == sorted(match)

    path = [{"time (inc)": ">= 30.0"}, ".", {"name": "bar"}, "*"]
    match = [
        root.children[1].children[0],
        root.children[1].children[0].children[0],
        root.children[1].children[0].children[0].children[0],
        root.children[1].children[0].children[0].children[0].children[0],
        root.children[1].children[0].children[0].children[0].children[1],
        # Old-style return value of apply
        # [
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0].children[0],
        # ],
        # [
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0].children[1],
        # ],
    ]
    query = QueryMatcher(path)
    assert sorted(query.apply(gf)) == sorted(match)

    path = [{"name": "foo"}, {"name": "bar"}, {"time": 5.0}]
    # match = [[root, root.children[0], root.children[0].children[0]]]
    match = [root, root.children[0], root.children[0].children[0]]
    query = QueryMatcher(path)
    assert sorted(query.apply(gf)) == sorted(match)

    path = [{"name": "foo"}, {"name": "qux"}, ("+", {"time (inc)": "> 15.0"})]
    match = [
        root,
        root.children[1],
        root.children[1].children[0],
        root.children[1].children[0].children[0],
        root.children[1].children[0].children[0].children[0],
        # Old-style return value of apply
        # [
        #     root,
        #     root.children[1],
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0],
        # ],
        # [
        #     root,
        #     root.children[1],
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        # ],
    ]
    query = QueryMatcher(path)
    assert sorted(query.apply(gf)) == sorted(match)

    path = [{"name": "this"}, ("*", {"name": "is"}), {"name": "nonsense"}]

    query = QueryMatcher(path)
    assert query.apply(gf) == []

    path = [{"name": 5}, "*", {"name": "whatever"}]
    query = QueryMatcher(path)
    with pytest.raises(InvalidQueryFilter):
        query.apply(gf)

    path = [{"time": "badstring"}, "*", {"name": "whatever"}]
    query = QueryMatcher(path)
    with pytest.raises(InvalidQueryFilter):
        query.apply(gf)

    class DummyType:
        def __init__(self):
            self.x = 5.0
            self.y = -1
            self.z = "hello"

    bad_field_test_dict = list(mock_graph_literal)
    bad_field_test_dict[0]["children"][0]["children"][0]["metrics"][
        "list"
    ] = DummyType()
    gf = GraphFrame.from_literal(bad_field_test_dict)
    path = [{"name": "foo"}, {"name": "bar"}, {"list": DummyType()}]
    query = QueryMatcher(path)
    with pytest.raises(InvalidQueryFilter):
        query.apply(gf)

    path = ["*", {"name": "bar"}, {"name": "grault"}, "*"]
    match = [
        [root, root.children[0], root.children[0].children[1]],
        [root.children[0], root.children[0].children[1]],
        [
            root,
            root.children[1],
            root.children[1].children[0],
            root.children[1].children[0].children[0],
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            root.children[1],
            root.children[1].children[0],
            root.children[1].children[0].children[0],
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            root.children[1].children[0],
            root.children[1].children[0].children[0],
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            root.children[1].children[0].children[0],
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            gf.graph.roots[1],
            gf.graph.roots[1].children[0],
            gf.graph.roots[1].children[0].children[1],
        ],
        [gf.graph.roots[1].children[0], gf.graph.roots[1].children[0].children[1]],
    ]
    match = list(set().union(*match))
    query = QueryMatcher(path)
    assert sorted(query.apply(gf)) == sorted(match)

    path = ["*", {"name": "bar"}, {"name": "grault"}, "+"]
    query = QueryMatcher(path)
    assert query.apply(gf) == []


def test_apply_indices(calc_pi_hpct_db):
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))
    main = gf.graph.roots[0].children[0]
    path = [
        {"name": "[0-9]*:?MPI_.*"},
        ("*", {"name": "^((?!MPID).)*"}),
        {"name": "[0-9]*:?MPID.*"},
    ]
    matches = [
        [
            main.children[0],
            main.children[0].children[0],
            main.children[0].children[0].children[0],
            main.children[0].children[0].children[0].children[0],
        ],
        [
            main.children[1],
            main.children[1].children[0],
            main.children[1].children[0].children[0],
        ],
    ]
    matches = list(set().union(*matches))
    query = QueryMatcher(path)
    assert sorted(query.apply(gf)) == sorted(matches)

    gf.drop_index_levels()
    assert query.apply(gf) == matches


def test_high_level_depth(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query = QueryMatcher([("*", {"depth": 1})])
    roots = gf.graph.roots
    # matches = [[c] for r in roots for c in r.children]
    matches = [c for r in roots for c in r.children]
    assert sorted(query.apply(gf)) == sorted(matches)

    query = QueryMatcher([("*", {"depth": "<= 2"})])
    matches = [
        [roots[0], roots[0].children[0], roots[0].children[0].children[0]],
        [roots[0].children[0], roots[0].children[0].children[0]],
        [roots[0].children[0].children[0]],
        [roots[0], roots[0].children[0], roots[0].children[0].children[1]],
        [roots[0].children[0], roots[0].children[0].children[1]],
        [roots[0].children[0].children[1]],
        [roots[0], roots[0].children[1], roots[0].children[1].children[0]],
        [roots[0].children[1], roots[0].children[1].children[0]],
        [roots[0].children[1].children[0]],
        [roots[0], roots[0].children[2], roots[0].children[2].children[0]],
        [roots[0].children[2], roots[0].children[2].children[0]],
        [roots[0].children[2].children[0]],
        [roots[0], roots[0].children[2], roots[0].children[2].children[1]],
        [roots[0].children[2], roots[0].children[2].children[1]],
        [roots[0].children[2].children[1]],
        [roots[1], roots[1].children[0], roots[1].children[0].children[0]],
        [roots[1].children[0], roots[1].children[0].children[0]],
        [roots[1].children[0].children[0]],
        [roots[1], roots[1].children[0], roots[1].children[0].children[1]],
        [roots[1].children[0], roots[1].children[0].children[1]],
        [roots[1].children[0].children[1]],
    ]
    matches = list(set().union(*matches))
    assert sorted(query.apply(gf)) == sorted(matches)

    with pytest.raises(InvalidQueryFilter):
        query = QueryMatcher([{"depth": "hello"}])
        query.apply(gf)


def test_high_level_hatchet_nid(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query = QueryMatcher([("*", {"node_id": ">= 20"})])
    root = gf.graph.roots[1]
    matches = [
        [root, root.children[0], root.children[0].children[0]],
        [root.children[0], root.children[0].children[0]],
        [root.children[0].children[0]],
        [root, root.children[0], root.children[0].children[1]],
        [root.children[0], root.children[0].children[1]],
        [root.children[0].children[1]],
    ]
    matches = list(set().union(*matches))
    assert sorted(query.apply(gf)) == sorted(matches)

    query = QueryMatcher([{"node_id": 0}])
    # assert query.apply(gf) == [[gf.graph.roots[0]]]
    assert query.apply(gf) == [gf.graph.roots[0]]

    with pytest.raises(InvalidQueryFilter):
        query = QueryMatcher([{"node_id": "hello"}])
        query.apply(gf)


def test_high_level_depth_index_levels(calc_pi_hpct_db):
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))
    root = gf.graph.roots[0]

    query = QueryMatcher([("*", {"depth": "<= 2"})])
    matches = [
        [root, root.children[0], root.children[0].children[0]],
        [root.children[0], root.children[0].children[0]],
        [root.children[0].children[0]],
        [root, root.children[0], root.children[0].children[1]],
        [root.children[0], root.children[0].children[1]],
        [root.children[0].children[1]],
    ]
    matches = list(set().union(*matches))
    assert sorted(query.apply(gf)) == sorted(matches)

    query = QueryMatcher([("*", {"depth": 0})])
    # matches = [[root]]
    matches = [root]
    assert query.apply(gf) == matches

    with pytest.raises(InvalidQueryFilter):
        query = QueryMatcher([{"depth": "hello"}])
        query.apply(gf)


def test_high_level_node_id_index_levels(calc_pi_hpct_db):
    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))
    root = gf.graph.roots[0]

    query = QueryMatcher([("*", {"node_id": "<= 2"})])
    matches = [
        [root, root.children[0]],
        [root.children[0]],
        [root, root.children[0], root.children[0].children[0]],
        [root.children[0], root.children[0].children[0]],
        [root.children[0].children[0]],
    ]
    matches = list(set().union(*matches))
    assert sorted(query.apply(gf)) == sorted(matches)

    query = QueryMatcher([("*", {"node_id": 0})])
    # matches = [[root]]
    matches = [root]
    assert query.apply(gf) == matches

    with pytest.raises(InvalidQueryFilter):
        query = QueryMatcher([{"node_id": "hello"}])
        query.apply(gf)


def test_high_level_multi_condition_one_attribute(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query = QueryMatcher([("*", {"time (inc)": [">= 20", "<= 60"]})])
    roots = gf.graph.roots
    matches = [
        [roots[0].children[0]],
        [
            roots[0].children[1],
            roots[0].children[1].children[0],
            roots[0].children[1].children[0].children[0],
            roots[0].children[1].children[0].children[0].children[0],
        ],
        [
            roots[0].children[1],
            roots[0].children[1].children[0],
            roots[0].children[1].children[0].children[0],
        ],
        [
            roots[0].children[1].children[0],
            roots[0].children[1].children[0].children[0],
            roots[0].children[1].children[0].children[0].children[0],
        ],
        [
            roots[0].children[1].children[0],
            roots[0].children[1].children[0].children[0],
        ],
        [
            roots[0].children[1].children[0].children[0],
            roots[0].children[1].children[0].children[0].children[0],
        ],
        [roots[0].children[1].children[0].children[0]],
        [roots[0].children[1].children[0].children[0].children[0]],
        [
            roots[0].children[2],
            roots[0].children[2].children[0],
            roots[0].children[2].children[0].children[1],
            roots[0].children[2].children[0].children[1].children[0],
        ],
        [roots[0].children[2], roots[0].children[2].children[0]],
        [roots[0].children[2]],
        [
            roots[0].children[2].children[0],
            roots[0].children[2].children[0].children[1],
            roots[0].children[2].children[0].children[1].children[0],
        ],
        [roots[0].children[2].children[0]],
        [
            roots[0].children[2].children[0].children[1],
            roots[0].children[2].children[0].children[1].children[0],
        ],
        [roots[0].children[2].children[0].children[1].children[0]],
        [roots[1], roots[1].children[0]],
        [roots[1].children[0]],
    ]
    matches = list(set().union(*matches))
    assert sorted(query.apply(gf)) == sorted(matches)


def test_query_matcher_is_abstract_query():
    assert issubclass(QueryMatcher, AbstractQuery)


def test_nary_query_is_abstract_query():
    assert issubclass(NaryQuery, AbstractQuery)


def test_and_query_is_nary_query():
    assert issubclass(AndQuery, NaryQuery)


def test_or_query_is_nary_query():
    assert issubclass(OrQuery, NaryQuery)


def test_xor_query_is_nary_query():
    assert issubclass(XorQuery, NaryQuery)


def test_nary_query_high_level_construction(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query1 = [("*", {"time (inc)": [">= 20", "<= 60"]})]
    query2 = [("*", {"time (inc)": ">= 60"})]
    q1_node = gf.graph.roots[0].children[1].children[0].children[0]
    q2_node = gf.graph.roots[0]
    compound_query = AndQuery(query1, query2)
    assert compound_query.subqueries[0].query_pattern[0][0] == "*"
    assert compound_query.subqueries[0].query_pattern[0][1](gf.dataframe.loc[q1_node])
    assert not compound_query.subqueries[0].query_pattern[0][1](
        gf.dataframe.loc[q2_node]
    )
    assert compound_query.subqueries[1].query_pattern[0][0] == "*"
    assert compound_query.subqueries[1].query_pattern[0][1](gf.dataframe.loc[q2_node])
    assert not compound_query.subqueries[1].query_pattern[0][1](
        gf.dataframe.loc[q1_node]
    )


def test_nary_query_low_level_construction(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query1 = QueryMatcher().match(
        "*", lambda x: x["time (inc)"] >= 20 and x["time (inc)"] <= 60
    )
    query2 = QueryMatcher().match("*", lambda x: x["time (inc)"] >= 60)
    q1_node = gf.graph.roots[0].children[1].children[0].children[0]
    q2_node = gf.graph.roots[0]
    compound_query = AndQuery(query1, query2)
    assert compound_query.subqueries[0].query_pattern[0][0] == "*"
    assert compound_query.subqueries[0].query_pattern[0][1](gf.dataframe.loc[q1_node])
    assert not compound_query.subqueries[0].query_pattern[0][1](
        gf.dataframe.loc[q2_node]
    )
    assert compound_query.subqueries[1].query_pattern[0][0] == "*"
    assert compound_query.subqueries[1].query_pattern[0][1](gf.dataframe.loc[q2_node])
    assert not compound_query.subqueries[1].query_pattern[0][1](
        gf.dataframe.loc[q1_node]
    )


def test_nary_query_mixed_level_construction(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query1 = [("*", {"time (inc)": [">= 20", "<= 60"]})]
    query2 = QueryMatcher().match("*", lambda x: x["time (inc)"] >= 60)
    q1_node = gf.graph.roots[0].children[1].children[0].children[0]
    q2_node = gf.graph.roots[0]
    compound_query = AndQuery(query1, query2)
    assert compound_query.subqueries[0].query_pattern[0][0] == "*"
    assert compound_query.subqueries[0].query_pattern[0][1](gf.dataframe.loc[q1_node])
    assert not compound_query.subqueries[0].query_pattern[0][1](
        gf.dataframe.loc[q2_node]
    )
    assert compound_query.subqueries[1].query_pattern[0][0] == "*"
    assert compound_query.subqueries[1].query_pattern[0][1](gf.dataframe.loc[q2_node])
    assert not compound_query.subqueries[1].query_pattern[0][1](
        gf.dataframe.loc[q1_node]
    )


def test_and_query(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query1 = [("*", {"time (inc)": [">= 20", "<= 60"]})]
    query2 = [("*", {"time (inc)": ">= 60"})]
    compound_query = AndQuery(query1, query2)
    roots = gf.graph.roots
    matches = [
        roots[0].children[1],
        roots[0].children[1].children[0],
    ]
    assert sorted(compound_query.apply(gf)) == sorted(matches)


def test_intersection_query(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query1 = [("*", {"time (inc)": [">= 20", "<= 60"]})]
    query2 = [("*", {"time (inc)": ">= 60"})]
    compound_query = IntersectionQuery(query1, query2)
    roots = gf.graph.roots
    matches = [
        roots[0].children[1],
        roots[0].children[1].children[0],
    ]
    assert sorted(compound_query.apply(gf)) == sorted(matches)


def test_or_query(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query1 = [("*", {"time (inc)": 5.0})]
    query2 = [("*", {"time (inc)": 10.0})]
    compound_query = OrQuery(query1, query2)
    roots = gf.graph.roots
    matches = [
        roots[0].children[0].children[0],
        roots[0].children[0].children[1],
        roots[0].children[1].children[0].children[0].children[0].children[0],
        roots[0].children[1].children[0].children[0].children[0].children[1],
        roots[0].children[1].children[0].children[0].children[1],
        roots[0].children[2].children[0].children[0],
        roots[0].children[2].children[0].children[1].children[0].children[0],
        roots[1].children[0].children[0],
        roots[1].children[0].children[1],
    ]
    assert sorted(compound_query.apply(gf)) == sorted(matches)


def test_union_query(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query1 = [("*", {"time (inc)": 5.0})]
    query2 = [("*", {"time (inc)": 10.0})]
    compound_query = UnionQuery(query1, query2)
    roots = gf.graph.roots
    matches = [
        roots[0].children[0].children[0],
        roots[0].children[0].children[1],
        roots[0].children[1].children[0].children[0].children[0].children[0],
        roots[0].children[1].children[0].children[0].children[0].children[1],
        roots[0].children[1].children[0].children[0].children[1],
        roots[0].children[2].children[0].children[0],
        roots[0].children[2].children[0].children[1].children[0].children[0],
        roots[1].children[0].children[0],
        roots[1].children[0].children[1],
    ]
    assert sorted(compound_query.apply(gf)) == sorted(matches)


def test_xor_query(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query1 = [("*", {"time (inc)": [">= 5.0", "<= 10.0"]})]
    query2 = [("*", {"time (inc)": 10.0})]
    compound_query = XorQuery(query1, query2)
    roots = gf.graph.roots
    matches = [
        roots[0].children[0].children[0],
        # roots[0].children[0].children[1],
        roots[0].children[1].children[0].children[0].children[0].children[0],
        # roots[0].children[1].children[0].children[0].children[0].children[1],
        # roots[0].children[1].children[0].children[0].children[1],
        roots[0].children[2].children[0].children[0],
        roots[0].children[2].children[0].children[1].children[0].children[0],
        roots[1].children[0].children[0],
        # roots[1].children[0].children[1],
    ]
    assert sorted(compound_query.apply(gf)) == sorted(matches)


def test_sym_diff_query(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    query1 = [("*", {"time (inc)": [">= 5.0", "<= 10.0"]})]
    query2 = [("*", {"time (inc)": 10.0})]
    compound_query = SymDifferenceQuery(query1, query2)
    roots = gf.graph.roots
    matches = [
        roots[0].children[0].children[0],
        # roots[0].children[0].children[1],
        roots[0].children[1].children[0].children[0].children[0].children[0],
        # roots[0].children[1].children[0].children[0].children[0].children[1],
        # roots[0].children[1].children[0].children[0].children[1],
        roots[0].children[2].children[0].children[0],
        roots[0].children[2].children[0].children[1].children[0].children[0],
        roots[1].children[0].children[0],
        # roots[1].children[0].children[1],
    ]
    assert sorted(compound_query.apply(gf)) == sorted(matches)


def test_construct_cypher_api():
    mock_node_mpi = {"name": "MPI_Bcast"}
    mock_node_ibv = {"name": "ibv_reg_mr"}
    mock_node_time_true = {"time (inc)": 0.1}
    mock_node_time_false = {"time (inc)": 0.001}
    # path1 = [{"name": "MPI_[_a-zA-Z]*"}, "*", {"name": "ibv[_a-zA-Z]*"}]
    path1 = """MATCH (p)->("*")->(q)
    WHERE p."name" STARTS WITH "MPI_" AND q."name" STARTS WITH "ibv"
    """
    # path2 = [{"name": "MPI_[_a-zA-Z]*"}, 2, {"name": "ibv[_a-zA-Z]*"}]
    path2 = """MATCH (p)->(2)->(q)
    WHERE p."name" STARTS WITH "MPI_" AND q."name" STARTS WITH "ibv"
    """
    # path3 = [
    #     {"name": "MPI_[_a-zA-Z]*"},
    #     ("+", {"time (inc)": ">= 0.1"}),
    #     {"name": "ibv[_a-zA-Z]*"},
    # ]
    path3 = """MATCH (p)->("+", a)->(q)
    WHERE p."name" STARTS WITH "MPI" AND a."time (inc)" >= 0.1 AND q."name" STARTS WITH "ibv"
    """
    # path4 = [
    #     {"name": "MPI_[_a-zA-Z]*"},
    #     (3, {"time (inc)": 0.1}),
    #     {"name": "ibv[_a-zA-Z]*"},
    # ]
    path4 = """MATCH (p)->(3, a)->(q)
    WHERE p."name" STARTS WITH "MPI" AND a."time (inc)" = 0.1 AND q."name" STARTS WITH "ibv"
    """
    query1 = CypherQuery(path1)
    query2 = CypherQuery(path2)
    query3 = CypherQuery(path3)
    query4 = CypherQuery(path4)

    assert query1.query_pattern[0][0] == "."
    assert query1.query_pattern[0][1](mock_node_mpi)
    assert not query1.query_pattern[0][1](mock_node_ibv)
    assert not query1.query_pattern[0][1](mock_node_time_true)
    assert query1.query_pattern[1][0] == "*"
    assert query1.query_pattern[1][1](mock_node_mpi)
    assert query1.query_pattern[1][1](mock_node_ibv)
    assert query1.query_pattern[1][1](mock_node_time_true)
    assert query1.query_pattern[1][1](mock_node_time_false)
    assert query1.query_pattern[2][0] == "."

    assert query2.query_pattern[0][0] == "."
    assert query2.query_pattern[1][0] == "."
    assert query2.query_pattern[1][1](mock_node_mpi)
    assert query2.query_pattern[1][1](mock_node_ibv)
    assert query2.query_pattern[1][1](mock_node_time_true)
    assert query2.query_pattern[1][1](mock_node_time_false)
    assert query2.query_pattern[2][0] == "."
    assert query2.query_pattern[2][1](mock_node_mpi)
    assert query2.query_pattern[2][1](mock_node_ibv)
    assert query2.query_pattern[2][1](mock_node_time_true)
    assert query2.query_pattern[2][1](mock_node_time_false)
    assert query2.query_pattern[3][0] == "."

    assert query3.query_pattern[0][0] == "."
    assert query3.query_pattern[1][0] == "+"
    assert not query3.query_pattern[1][1](mock_node_mpi)
    assert not query3.query_pattern[1][1](mock_node_ibv)
    assert query3.query_pattern[1][1](mock_node_time_true)
    assert not query3.query_pattern[1][1](mock_node_time_false)
    assert query3.query_pattern[2][0] == "."

    assert query4.query_pattern[0][0] == "."
    assert query4.query_pattern[1][0] == "."
    assert not query4.query_pattern[1][1](mock_node_mpi)
    assert not query4.query_pattern[1][1](mock_node_ibv)
    assert query4.query_pattern[1][1](mock_node_time_true)
    assert not query4.query_pattern[1][1](mock_node_time_false)
    assert query4.query_pattern[2][0] == "."
    assert not query4.query_pattern[2][1](mock_node_mpi)
    assert not query4.query_pattern[2][1](mock_node_ibv)
    assert query4.query_pattern[2][1](mock_node_time_true)
    assert not query4.query_pattern[2][1](mock_node_time_false)
    assert query4.query_pattern[3][0] == "."
    assert not query4.query_pattern[3][1](mock_node_mpi)
    assert not query4.query_pattern[3][1](mock_node_ibv)
    assert query4.query_pattern[3][1](mock_node_time_true)
    assert not query4.query_pattern[3][1](mock_node_time_false)
    assert query4.query_pattern[4][0] == "."

    # invalid_path = [
    #     {"name": "MPI_[_a-zA-Z]*"},
    #     ({"bad": "wildcard"}, {"time (inc)": 0.1}),
    #     {"name": "ibv[_a-zA-Z]*"},
    # ]
    invalid_path = """MATCH (p)->({"bad": "wildcard"}, a)->(q)
    WHERE p."name" STARTS WITH "MPI" AND a."time (inc)" = 0.1 AND
    q."name" STARTS WITH "ibv"
    """
    with pytest.raises(InvalidQueryPath):
        _ = CypherQuery(invalid_path)


def test_apply_cypher(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    # path = [
    #     {"time (inc)": ">= 30.0"},
    #     (2, {"name": "[^b][a-z]+"}),
    #     ("*", {"name": "[^b][a-z]+"}),
    #     {"name": "gr[a-z]+"},
    # ]
    path = """MATCH (p)->(2, q)->("*", r)->(s)
    WHERE p."time (inc)" >= 30.0 AND NOT q."name" STARTS WITH "b"
    AND r."name" =~ "[^b][a-z]+" AND s."name" STARTS WITH "gr"
    """
    root = gf.graph.roots[0]
    match = [
        root,
        root.children[1],
        root.children[1].children[0],
        root.children[1].children[0].children[0],
        root.children[1].children[0].children[0].children[1],
        # Old-style return value of apply
        # [
        #     root,
        #     root.children[1],
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[1],
        # ],
        # [
        #     root.children[1],
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[1],
        # ],
    ]
    query = CypherQuery(path)

    assert sorted(query.apply(gf)) == sorted(match)

    # path = [{"time (inc)": ">= 30.0"}, ".", {"name": "bar"}, "*"]
    path = """MATCH (p)->(".")->(q)->("*")
    WHERE p."time (inc)" >= 30.0 AND q."name" = "bar"
    """
    match = [
        root.children[1].children[0],
        root.children[1].children[0].children[0],
        root.children[1].children[0].children[0].children[0],
        root.children[1].children[0].children[0].children[0].children[0],
        root.children[1].children[0].children[0].children[0].children[1],
        # Old-style return value of apply
        # [
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0].children[0],
        # ],
        # [
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0].children[1],
        # ],
    ]
    query = CypherQuery(path)
    assert sorted(query.apply(gf)) == sorted(match)

    # path = [{"name": "foo"}, {"name": "bar"}, {"time": 5.0}]
    path = """MATCH (p)->(q)->(r)
    WHERE p."name" = "foo" AND q."name" = "bar" AND r."time" = 5.0
    """
    # match = [[root, root.children[0], root.children[0].children[0]]]
    match = [root, root.children[0], root.children[0].children[0]]
    query = CypherQuery(path)
    assert sorted(query.apply(gf)) == sorted(match)

    # path = [{"name": "foo"}, {"name": "qux"}, ("+", {"time (inc)": "> 15.0"})]
    path = """MATCH (p)->(q)->("+", r)
    WHERE p."name" = "foo" AND q."name" = "qux" AND r."time (inc)" > 15.0
    """
    match = [
        root,
        root.children[1],
        root.children[1].children[0],
        root.children[1].children[0].children[0],
        root.children[1].children[0].children[0].children[0],
        # Old-style return value of apply
        # [
        #     root,
        #     root.children[1],
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        #     root.children[1].children[0].children[0].children[0],
        # ],
        # [
        #     root,
        #     root.children[1],
        #     root.children[1].children[0],
        #     root.children[1].children[0].children[0],
        # ],
    ]
    query = CypherQuery(path)
    assert sorted(query.apply(gf)) == sorted(match)

    path = """MATCH (p)->(q)
    WHERE p."time (inc)" > 100 OR p."time (inc)" <= 30 AND q."time (inc)" = 20
    """
    roots = gf.graph.roots
    match = [
        roots[0],
        roots[0].children[0],
        roots[1],
        roots[1].children[0],
    ]
    query = CypherQuery(path)
    assert sorted(query.apply(gf)) == sorted(match)

    # path = [{"name": "this"}, ("*", {"name": "is"}), {"name": "nonsense"}]
    path = """MATCH (p)->("*", q)->(r)
    WHERE p."name" = "this" AND q."name" = "is" AND r."name" = "nonsense"
    """

    query = CypherQuery(path)
    assert query.apply(gf) == []

    # path = [{"name": 5}, "*", {"name": "whatever"}]
    path = """MATCH (p)->("*")->(q)
    WHERE p."name" = 5 AND q."name" = "whatever"
    """
    with pytest.raises(InvalidQueryFilter):
        query = CypherQuery(path)
        query.apply(gf)

    # path = [{"time": "badstring"}, "*", {"name": "whatever"}]
    path = """MATCH (p)->("*")->(q)
    WHERE p."time" = "badstring" AND q."name" = "whatever"
    """
    query = CypherQuery(path)
    with pytest.raises(InvalidQueryFilter):
        query.apply(gf)

    class DummyType:
        def __init__(self):
            self.x = 5.0
            self.y = -1
            self.z = "hello"

    bad_field_test_dict = list(mock_graph_literal)
    bad_field_test_dict[0]["children"][0]["children"][0]["metrics"][
        "list"
    ] = DummyType()
    gf = GraphFrame.from_literal(bad_field_test_dict)
    # path = [{"name": "foo"}, {"name": "bar"}, {"list": DummyType()}]
    path = """MATCH (p)->(q)->(r)
    WHERE p."name" = "foo" AND q."name" = "bar" AND p."list" = DummyType()
    """
    with pytest.raises(InvalidQueryPath):
        query = CypherQuery(path)
        query.apply(gf)

    # path = ["*", {"name": "bar"}, {"name": "grault"}, "*"]
    path = """MATCH ("*")->(p)->(q)->("*")
    WHERE p."name" = "bar" AND q."name" = "grault"
    """
    match = [
        [root, root.children[0], root.children[0].children[1]],
        [root.children[0], root.children[0].children[1]],
        [
            root,
            root.children[1],
            root.children[1].children[0],
            root.children[1].children[0].children[0],
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            root.children[1],
            root.children[1].children[0],
            root.children[1].children[0].children[0],
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            root.children[1].children[0],
            root.children[1].children[0].children[0],
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            root.children[1].children[0].children[0],
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            root.children[1].children[0].children[0].children[0],
            root.children[1].children[0].children[0].children[0].children[1],
        ],
        [
            gf.graph.roots[1],
            gf.graph.roots[1].children[0],
            gf.graph.roots[1].children[0].children[1],
        ],
        [gf.graph.roots[1].children[0], gf.graph.roots[1].children[0].children[1]],
    ]
    match = list(set().union(*match))
    query = CypherQuery(path)
    assert sorted(query.apply(gf)) == sorted(match)

    # path = ["*", {"name": "bar"}, {"name": "grault"}, "+"]
    path = """MATCH ("*")->(p)->(q)->("+")
    WHERE p."name" = "bar" AND q."name" = "grault"
    """
    query = CypherQuery(path)
    assert query.apply(gf) == []

    gf.dataframe["time"] = np.NaN
    gf.dataframe.at[gf.graph.roots[0], "time"] = 5.0
    path = """MATCH ("*", p)
    WHERE p."time" IS NOT NAN"""
    match = [gf.graph.roots[0]]
    query = CypherQuery(path)
    assert query.apply(gf) == match

    gf.dataframe["time"] = 5.0
    gf.dataframe.at[gf.graph.roots[0], "time"] = np.NaN
    path = """MATCH ("*", p)
    WHERE p."time" IS NAN"""
    match = [gf.graph.roots[0]]
    query = CypherQuery(path)
    assert query.apply(gf) == match

    gf.dataframe["time"] = np.Inf
    gf.dataframe.at[gf.graph.roots[0], "time"] = 5.0
    path = """MATCH ("*", p)
    WHERE p."time" IS NOT INF"""
    match = [gf.graph.roots[0]]
    query = CypherQuery(path)
    assert query.apply(gf) == match

    gf.dataframe["time"] = 5.0
    gf.dataframe.at[gf.graph.roots[0], "time"] = np.Inf
    path = """MATCH ("*", p)
    WHERE p."time" IS INF"""
    match = [gf.graph.roots[0]]
    query = CypherQuery(path)
    assert query.apply(gf) == match

    names = gf.dataframe["name"].copy()
    gf.dataframe["name"] = None
    gf.dataframe.at[gf.graph.roots[0], "name"] = names.iloc[0]
    path = """MATCH ("*", p)
    WHERE p."name" IS NOT NONE"""
    match = [gf.graph.roots[0]]
    query = CypherQuery(path)
    assert query.apply(gf) == match

    gf.dataframe["name"] = names
    gf.dataframe.at[gf.graph.roots[0], "name"] = None
    path = """MATCH ("*", p)
    WHERE p."name" IS NONE"""
    match = [gf.graph.roots[0]]
    query = CypherQuery(path)
    assert query.apply(gf) == match



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json
import sys
import re
import subprocess
import os
import math

import pandas as pd
import numpy as np

import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame
from hatchet.util.timer import Timer
from hatchet.util.executable import which

unknown_label_counter = 0


class CaliperReader:
    """Read in a Caliper file (`cali` or split JSON) or file-like object."""

    def __init__(self, filename_or_stream, query=""):
        """Read from Caliper files (`cali` or split JSON).

        Args:
            filename_or_stream (str or file-like): name of a `cali` or
                `cali-query` split JSON file, OR an open file object
            query (str): cali-query arguments (for cali file)
        """
        self.filename_or_stream = filename_or_stream
        self.filename_ext = ""
        self.query = query

        self.json_data = {}
        self.json_cols = {}
        self.json_cols_mdata = {}
        self.json_nodes = {}

        self.metadata = {}

        self.idx_to_label = {}
        self.idx_to_node = {}

        self.timer = Timer()
        self.nid_col_name = "nid"

        if isinstance(self.filename_or_stream, str):
            _, self.filename_ext = os.path.splitext(filename_or_stream)

    def read_json_sections(self):
        # if cali-query exists, extract data from .cali to a file-like object
        if self.filename_ext == ".cali":
            cali_query = which("cali-query")
            if not cali_query:
                raise ValueError("from_caliper() needs cali-query to query .cali file")
            cali_json = subprocess.Popen(
                [cali_query, "-q", self.query, self.filename_or_stream],
                stdout=subprocess.PIPE,
            )
            self.filename_or_stream = cali_json.stdout

        # if filename_or_stream is a str, then open the file, otherwise
        # directly load the file-like object
        if isinstance(self.filename_or_stream, str):
            with open(self.filename_or_stream) as cali_json:
                json_obj = json.load(cali_json)
        else:
            json_obj = json.loads(self.filename_or_stream.read().decode("utf-8"))

        # read various sections of the Caliper JSON file
        self.json_data = json_obj["data"]
        self.json_cols = json_obj["columns"]
        self.json_cols_mdata = json_obj["column_metadata"]
        self.json_nodes = json_obj["nodes"]

        # read run metadata: all top-level elements in the json object that aren't
        # one of the above sections are run metadata
        skip = ["data", "columns", "column_metadata", "nodes"]
        keys = [k for k in json_obj.keys() if k not in skip]
        self.metadata = {k: json_obj[k] for k in keys}

        # decide which column to use as the primary path hierarchy
        # first preference to callpath if available
        if "source.function#callpath.address" in self.json_cols:
            self.path_col_name = "source.function#callpath.address"
            self.node_type = "function"
            if "path" in self.json_cols:
                self.both_hierarchies = True
            else:
                self.both_hierarchies = False
        elif "path" in self.json_cols:
            self.path_col_name = "path"
            self.node_type = "region"
            self.both_hierarchies = False
        else:
            sys.exit("No hierarchy column in input file")

        # remove data entries containing None in `path` column (null in json file)
        # first, get column where `path` data is
        # then, parse json_data list of lists to identify lists containing None in
        # `path` column
        path_col = self.json_cols.index(self.path_col_name)
        entries_to_remove = []
        for sublist in self.json_data:
            if sublist[path_col] is None:
                entries_to_remove.append(sublist)
        # then, remove them from the json_data list
        for i in entries_to_remove:
            self.json_data.remove(i)

        # change column names
        for idx, item in enumerate(self.json_cols):
            if item == self.path_col_name:
                # this column is just a pointer into the nodes section
                self.json_cols[idx] = self.nid_col_name
            # make other columns consistent with other readers
            if item == "mpi.rank":
                self.json_cols[idx] = "rank"
            if item == "module#cali.sampler.pc":
                self.json_cols[idx] = "module"
            if item == "sum#time.duration" or item == "sum#avg#sum#time.duration":
                self.json_cols[idx] = "time"
            if (
                item == "inclusive#sum#time.duration"
                or item == "sum#avg#inclusive#sum#time.duration"
            ):
                self.json_cols[idx] = "time (inc)"

        # make list of metric columns
        self.metric_columns = []
        for idx, item in enumerate(self.json_cols_mdata):
            if self.json_cols[idx] != "rank" and item["is_value"] is True:
                self.metric_columns.append(self.json_cols[idx])

    def create_graph(self):
        list_roots = []

        global unknown_label_counter

        # find nodes in the nodes section that represent the path hierarchy
        for idx, node in enumerate(self.json_nodes):
            node_label = node["label"]
            if node_label == "":
                node_label = "UNKNOWN " + str(unknown_label_counter)
                unknown_label_counter += 1
            self.idx_to_label[idx] = node_label

            if node["column"] == self.path_col_name:
                if "parent" not in node:
                    # since this node does not have a parent, this is a root
                    graph_root = Node(
                        Frame({"type": self.node_type, "name": node_label}), None
                    )
                    list_roots.append(graph_root)

                    node_dict = {
                        self.nid_col_name: idx,
                        "name": node_label,
                        "node": graph_root,
                    }
                    self.idx_to_node[idx] = node_dict
                else:
                    parent_hnode = (self.idx_to_node[node["parent"]])["node"]
                    hnode = Node(
                        Frame({"type": self.node_type, "name": node_label}),
                        parent_hnode,
                    )
                    parent_hnode.add_child(hnode)

                    node_dict = {
                        self.nid_col_name: idx,
                        "name": node_label,
                        "node": hnode,
                    }
                    self.idx_to_node[idx] = node_dict

        return list_roots

    def read(self):
        """Read the caliper JSON file to extract the calling context tree."""
        with self.timer.phase("read json"):
            self.read_json_sections()

        with self.timer.phase("graph construction"):
            list_roots = self.create_graph()

        # create a dataframe of metrics from the data section
        self.df_json_data = pd.DataFrame(self.json_data, columns=self.json_cols)

        # when an nid has multiple entries due to the secondary hierarchy
        # we need to aggregate them for each (nid, rank)
        groupby_cols = [self.nid_col_name]
        if "rank" in self.json_cols:
            groupby_cols.append("rank")
        if "sourceloc#cali.sampler.pc" in self.json_cols:
            groupby_cols.append("sourceloc#cali.sampler.pc")

        if self.both_hierarchies is True:
            # create dict that stores aggregation function for each column
            agg_dict = {}
            for idx, item in enumerate(self.json_cols_mdata):
                col = self.json_cols[idx]
                if col != "rank" and col != "nid":
                    if item["is_value"] is True:
                        agg_dict[col] = np.sum
                    else:
                        agg_dict[col] = lambda x: x.iloc[0]

            grouped = self.df_json_data.groupby(groupby_cols).aggregate(agg_dict)
            self.df_json_data = grouped.reset_index()

        # map non-numeric columns to their mappings in the nodes section
        for idx, item in enumerate(self.json_cols_mdata):
            if item["is_value"] is False and self.json_cols[idx] != self.nid_col_name:
                if self.json_cols[idx] == "sourceloc#cali.sampler.pc":
                    # split source file and line number into two columns
                    self.df_json_data["file"] = self.df_json_data[
                        self.json_cols[idx]
                    ].apply(
                        lambda x: re.match(
                            r"(.*):(\d+)", self.json_nodes[x]["label"]
                        ).group(1)
                    )
                    self.df_json_data["line"] = self.df_json_data[
                        self.json_cols[idx]
                    ].apply(
                        lambda x: re.match(
                            r"(.*):(\d+)", self.json_nodes[x]["label"]
                        ).group(2)
                    )
                    self.df_json_data.drop(self.json_cols[idx], axis=1, inplace=True)
                    sourceloc_idx = idx
                elif self.json_cols[idx] == "path":
                    # we will only reach here if path is the "secondary"
                    # hierarchy in the data
                    self.df_json_data["path"] = self.df_json_data["path"].apply(
                        lambda x: (
                            None
                            if (math.isnan(x))
                            else self.json_nodes[int(x)]["label"]
                        )
                    )
                else:
                    self.df_json_data[self.json_cols[idx]] = self.df_json_data[
                        self.json_cols[idx]
                    ].apply(lambda x: self.json_nodes[x]["label"])

        # since we split sourceloc, we should update json_cols and
        # json_cols_mdata
        if "sourceloc#cali.sampler.pc" in self.json_cols:
            self.json_cols.pop(sourceloc_idx)
            self.json_cols_mdata.pop(sourceloc_idx)
            self.json_cols.append("file")
            self.json_cols.append("line")
            self.json_cols_mdata.append({"is_value": False})
            self.json_cols_mdata.append({"is_value": False})

        max_nid = self.df_json_data[self.nid_col_name].max()

        if "line" in self.df_json_data.columns:
            # split nodes that have multiple file:line numbers to have a child
            # each with a unique file:line number
            unique_nodes = self.df_json_data.groupby(self.nid_col_name)
            df_concat = [self.df_json_data]

            for nid, super_node in unique_nodes:
                line_groups = super_node.groupby("line")
                # only need to do something if there are more than one
                # file:line number entries for the node
                if len(line_groups.size()) > 1:
                    sn_hnode = self.idx_to_node[nid]["node"]

                    for line, line_group in line_groups:
                        # create the node label
                        file_path = (line_group.head(1))["file"].item()
                        file_name = os.path.basename(file_path)
                        node_label = file_name + ":" + line

                        # create a new hatchet node
                        max_nid += 1
                        idx = max_nid
                        hnode = Node(
                            Frame(
                                {"type": "statement", "file": file_path, "line": line}
                            ),
                            sn_hnode,
                        )
                        sn_hnode.add_child(hnode)

                        node_dict = {
                            self.nid_col_name: idx,
                            "name": node_label,
                            "node": hnode,
                        }
                        self.idx_to_node[idx] = node_dict

                        # change nid of the original node to new node in place
                        for index, row in line_group.iterrows():
                            self.df_json_data.loc[index, "nid"] = max_nid

                    # add new row for original node
                    node_copy = super_node.head(1).copy()
                    for cols in self.metric_columns:
                        node_copy[cols] = 0
                    df_concat.append(node_copy)

            # concatenate all the newly created dataframes with
            # self.df_json_data
            self.df_fixed_data = pd.concat(df_concat)
        else:
            self.df_fixed_data = self.df_json_data

        # create a dataframe with all nodes in the call graph
        self.df_nodes = pd.DataFrame.from_dict(data=list(self.idx_to_node.values()))

        # add missing intermediate nodes to the df_fixed_data dataframe
        if "rank" in self.json_cols:
            self.num_ranks = self.df_fixed_data["rank"].max() + 1
            rank_list = range(0, self.num_ranks)

        # create a standard dict to be used for filling all missing rows
        default_metric_dict = {}
        for idx, item in enumerate(self.json_cols_mdata):
            if self.json_cols[idx] != self.nid_col_name:
                if item["is_value"] is True:
                    default_metric_dict[self.json_cols[idx]] = 0
                else:
                    default_metric_dict[self.json_cols[idx]] = None

        # create a list of dicts, one dict for each missing row
        missing_nodes = []
        for iteridx, row in self.df_nodes.iterrows():
            # check if df_nodes row exists in df_fixed_data
            metric_rows = self.df_fixed_data.loc[
                self.df_fixed_data[self.nid_col_name] == row[self.nid_col_name]
            ]
            if "rank" not in self.json_cols:
                if metric_rows.empty:
                    # add a single row
                    node_dict = dict(default_metric_dict)
                    node_dict[self.nid_col_name] = row[self.nid_col_name]
                    missing_nodes.append(node_dict)
            else:
                if metric_rows.empty:
                    # add a row per MPI rank
                    for rank in rank_list:
                        node_dict = dict(default_metric_dict)
                        node_dict[self.nid_col_name] = row[self.nid_col_name]
                        node_dict["rank"] = rank
                        missing_nodes.append(node_dict)
                elif len(metric_rows) < self.num_ranks:
                    # add a row for each missing MPI rank
                    present_ranks = metric_rows["rank"].values
                    missing_ranks = [x for x in rank_list if x not in present_ranks]
                    for rank in missing_ranks:
                        node_dict = dict(default_metric_dict)
                        node_dict[self.nid_col_name] = row[self.nid_col_name]
                        node_dict["rank"] = rank
                        missing_nodes.append(node_dict)

        self.df_missing = pd.DataFrame.from_dict(data=missing_nodes)
        self.df_metrics = pd.concat([self.df_fixed_data, self.df_missing])

        # create a graph object once all the nodes have been added
        graph = Graph(list_roots)
        graph.enumerate_traverse()

        # merge the metrics and node dataframes on the idx column
        with self.timer.phase("data frame"):
            dataframe = pd.merge(self.df_metrics, self.df_nodes, on=self.nid_col_name)
            # set the index to be a MultiIndex
            indices = ["node"]
            if "rank" in self.json_cols:
                indices.append("rank")
            dataframe.set_index(indices, inplace=True)
            dataframe.sort_index(inplace=True)

        # create list of exclusive and inclusive metric columns
        exc_metrics = []
        inc_metrics = []
        for column in self.metric_columns:
            if "(inc)" in column:
                inc_metrics.append(column)
            else:
                exc_metrics.append(column)

        return hatchet.graphframe.GraphFrame(
            graph, dataframe, exc_metrics, inc_metrics, metadata=self.metadata
        )



# Copyright 2023-2024 Advanced Micro Devices, Inc., and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd

from ..graph import Graph


def graphframe_indexing_helper(
    root_nodes,
    data,
    extensions=["rank", "thread"],
    columns=[],
    fill_value=0,
    fill_missing_attribute={},
    deepcopy=True,
):
    """Helper for the complicated scenario of reindexing the dataframe
    to a multi-index for dataframes with multiple ranks and/or threads.

    In every readers, the implementation boils down to the creation of
    two principal data structures:

        1. list of root nodes
        2. list of dicts containing table data: node, rank, thread, time, time (inc), etc.

    once these are created, the implementation varies from reader to reader
    but it all boils down to:

        1. create Graph from root nodes + enumerate_traverse()
        2. create dataframe, set index, sort index

    However, when multiple ranks and/or threads are introduced, a MultiIndex
    needs to be used but creating a MultiIndex is problematic because it
    essentially results in pandas giving each rank/thread/rank+thread the
    same call-stack, e.g. if rank 0 called 'foo' and rank 1 called 'bar'
    and A in Node({name: foo}) and B is Node({name: bar}):

        node    rank    name    time
        A       0       foo     1.0
        B       1       bar     0.9

    then a multi-index for ["node", "rank"] requires A+0, A+1, B+0, B+1
    so it fills in B+0 and A+1 since only A+0 and B+1 exist in the
    original data. Thus you end up with a data frame that looks like this:

        node    rank    name    time
        A       0       foo     1.0
        B       0       NaN     NaN
        A       1       NaN     NaN
        B       1       bar     0.9

    Only the columns in the index are copied so we need to "fill" in the
    relevant data to these new entries. We do this by taking the data
    encoded in the dictionary of the node column and copying it over
    into the columns with the same name.
    Since, A+1 and B+0 did not actually exist in the trace/profile,
    we can ignore columns such as "time" so we set it to zero, however,
    the name column is important bc it identifies the name of the function.
    We do that by copying the data encoded into the dict within the node
    column. I.e. since A+1 was created from A+0, it copied over the
    Node({name: foo}) and we can essentially want to do:

    for row in df.rows:
        row["name"] = row["node"].frame["name"]

    (i.e. dict {name: foo} is stored in Node.frame)

    Performing this operation is a bit difficult to perform efficiently
    because the most intuitive was (with dataframe.iterrows) is extremely
    slow. Thus we provide this helper function.

    Args:
        root_nodes (list of hatchet.Node):
            root nodes

        data (list/tuple of dict or pandas.dataframe):
            dict entries are {column_name: value}

        extensions (list of str):
            possible multi-index column names. "node" column is required by
            hatchet, "rank" and "thread" columns are optional. It is safe to
            use the default value even if you do not have these columns:
            this function checks if the columns even exist and also checks to
            make sure there is more than 1 unique value -- in other words, if
            you have both "rank" and "thread" columns but every row has a
            value of 0 and 0, then it is not necessary to create a multi-index
            and this function recognizes that. Similarly, if the data is
            single rank but multiple threads, this function will only create
            a multi-index on "node" and "thread" since factoring in
            "rank" to the multi-index is unnecessary

        columns (list of str):
            This is the dict keys from Node.frame whose values should be copied
            over into the dataframe columns. When set to an empty list, it
            will take the keys from the first Node.frame and use those. Thus,
            EVERY Node.frame SHOULD HAVE THE SAME SET OF KEYS. If, for example,
            sometimes you have file information for a node but not for others,
            set Node.frame["file"] = "<unknown>" or whatever you prefer for the
            value, just make sure the key exists and is neither None nor NaN.

        fill_value (number):
            In the case of multi-index, the NaN's need to be replaced with some
            value, zero is recommended.

        fill_missing_attribute (dict):
            If Node.frame does not have attribute with a specific key, fill with
            the given value instead of using fill_value. E.g.,
            if columns=["name", "file", "line"] and fill_value=0 and the
            Node.frame at a given index only has a "name" attribute, and you
            want the "file" column to be set to "<unknown>" instead of
            0 and the "line" column to be set to -1 then pass the argument:

                fill_missing_attribute={ "file": "<unknown>", "line": -1 }

        deep_copy (bool):
            Perform a full deep-copy of the dataframe (if data argument is
            dataframe). If you are not using the dataframe afterwards or
            assigning the returned dataframe to , set
            to False.
    """

    # make a copy since we cannot update argument inplace
    if isinstance(data, pd.DataFrame):
        if data.empty:
            raise RuntimeError("dataframe is empty")
        df_value = data.copy(deep=deepcopy)
    elif isinstance(data, (tuple, list)):
        if not data:
            raise RuntimeError("data is empty")
        elif not isinstance(data[0], dict):
            raise RuntimeError("data in list/tuple is not a dict")
        df_value = pd.DataFrame(data=list(data))
    else:
        raise TypeError("data type must be: pandas.DataFrame, list/tuple of dicts")

    graph_value = Graph(root_nodes)
    graph_value.enumerate_traverse()

    if df_value.empty:
        raise ValueError("dataframe is empty")

    indices = ["node"]
    for ext in extensions:
        if ext in df_value:
            n = len(df_value[ext].unique())
            if n > 1:
                indices.append(ext)

    # if there is no need for a multi-index: set index, sort, and return
    if len(indices) == 1:
        df_value.set_index(indices, inplace=True)
        df_value.sort_index(inplace=True)
        return (graph_value, df_value)

    new_index = pd.MultiIndex.from_product(
        [df_value[itr].unique() for itr in indices],
        names=indices,
    )
    df_value.set_index(indices, inplace=True)
    df_value = df_value.reindex(new_index, fill_value=fill_value)
    df_value.sort_index(inplace=True)

    # set columns to the list of columns to pull from Node.frame and
    # propagate to a column
    #
    # if no additional columns should be copied (aside from name), set columns=None
    # otherwise, if columns is empty list, find first Node().frame and extract the keys
    # not named "node"
    if not columns and columns is not None:
        columns = df_value.first_valid_index()[0].frame.attrs.keys()

    if columns is not None:
        # filter out "node" from columns
        columns = [x for x in columns if x not in ["node"]]
        # make sure name is copy cols
        if "name" not in columns:
            columns.append("name")
    else:
        columns = ["name"]

    # Select rows where name is 0, we need to fix that.
    zeros = df_value.loc[df_value["name"] == fill_value]

    # get a list of the indices
    zeros_idx = zeros.index.to_list()
    # make a list of lists where each inner list (columns) are:
    #   0  -> the indices
    #   1+ -> Node.frame values for the columns we are copying
    # we will zip together the outer list to enable iterating over
    # data
    zeros_data = [zeros_idx] + [
        [
            # extract the value if it exists, otherwise check fill_missing_attribute
            # for a potential override of fill_value
            itr[0].frame.get(col, fill_missing_attribute.get(col, fill_value))
            for itr in zeros_idx
        ]
        for col in columns
    ]

    # put values from Node.frame dictionary into column
    for ditr in zip(*zeros_data):
        # we placed the index at pos 0 and then
        # all the columns we are copying are in the same order
        # as columns but offset by 1 (because 0 is index)
        index = ditr[0]
        for idx, col in enumerate(columns):
            if df_value.at[index, col] == fill_value:
                df_value.at[index, col] = ditr[idx + 1]

    return (graph_value, df_value)



# Copyright 2023-2024 Advanced Micro Devices, Inc., and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import re
import pandas as pd

from collections import OrderedDict
from hatchet.graphframe import GraphFrame
from perfetto.trace_processor import TraceProcessor

from ..node import Node
from ..frame import Frame
from ..util.timer import Timer
from ..util.readers import graphframe_indexing_helper


class PerfettoReader:
    """Read in perfetto protobuf output"""

    def __init__(self, filename, select=None, **kwargs):
        """Arguments:
        filename (str or list/tuple of str):
            Valid arguments should be a list of files

        verbose (int):
            Information about processes, threads, categories, performance, etc.

        report (list of str):
            alternative for verbose. Accepts:

            - category (report categories in files)
            - process (report process info in files)
            - threads (report thread info in files)
            - track_ids (report track id info in files)
            - profile (report timing info from processing)

        exclude_category (list of str):
            Every slice has an associated category, e.g., slices from sampling may be
            in "sampling" category, slices from instrumentation may be in "instrumentation"
            category. The categories in protobuf is specific to the tool that generated the
            protobuf. Use this option to exclude the slices from specific categories

        include_category (list of str):
            Every slice has an associated category, e.g., slices from sampling may be
            in "sampling" category, slices from instrumentation may be in "instrumentation"
            category. The categories in protobuf is specific to the tool that generated the
            protobuf. Use this option to restrict the slices to the specified categories

        default_categories (list of str):
            Use this option as a safety value when include/exclude are used. These categories
            are used when include/exclude category filtering resulted in no data in the data
            frame. Accepts "all" if you want to fallback to just including all categories.

        patterns (list of regex str):
            Use this option to specify how the function/file/line are extracted from labels.
            For example, the default patterns are:

                r"(?P<func>.*) \\[(?P<file>\\S+):(?P<line>[0-9]+)\\]$"
                r"(?P<func>.*) \\[(?P<file>\\S+)\\]$"
                r"^(?P<file>\\S+):(?P<line>[0-9]+)$"

            To extract the function/file/line from the labeling patterns:

                func [file:line]
                func [file]
                file:line

            respectively.

        thread_index_regex (regex str):
            Use this option to determine thread IDs from the name of the threads. For example,
            Omnitrace labels certain process tracks "Thread X (S)" to indicate that the given
            process track are samples of Thread X. The default pattern is:

                r"(T|t)hread (?P<thread_index>[0-9]+)( |$)"

            to extract the "thread_index" field.

        max_depth (int):
            Set this a positive non-zero number signifying the maximum call-stack depth to
            process. This can significantly reduce the processing time for large traces.
        """

        self.timer = Timer()
        self.filename = filename if isinstance(filename, (list, tuple)) else [filename]
        self.metadata = {"hatchet_inclusive_suffix": ".inc"}
        self.default_metric = "time{}".format(self.metadata["hatchet_inclusive_suffix"])
        self.verbose = 0
        self.report = []
        self.exclude = []
        self.include = []
        self.categories = []
        self.default_categories = []
        self.df_categories = []
        self.dataframe = pd.DataFrame()
        self.process = pd.DataFrame()
        self.threads = pd.DataFrame()
        self.track_ids = []
        self.trace_processor = []
        self.compiled_patterns = []
        self.thread_index_regex = None
        self.max_depth = None
        self.configure(**kwargs)

    def configure(self, **kwargs):
        self.timer.start_phase("PerfettoReader config")

        # pre-compile the regex patterns for extracting the func, file, and line info
        # users can use their own pattens via patterns=[...]. An empty set of patterns
        # is valid for avoiding parsing func/file/line info
        _default_patterns = [
            # func [file:line]
            r"(?P<func>.*) \[(?P<file>\S+):(?P<line>[0-9]+)\]$",
            # func [file]
            r"(?P<func>.*) \[(?P<file>\S+)\]$",
            # file:line
            r"^(?P<file>\S+):(?P<line>[0-9]+)$",
        ]
        _patterns = kwargs["patterns"] if "patterns" in kwargs else None

        if _patterns is None:
            _patterns = _default_patterns
        elif kwargs.get("use_default_patterns", True):
            _patterns += _default_patterns

        self.compiled_patterns = [re.compile(x) for x in _patterns]
        self.thread_index_regex = re.compile(
            kwargs.get("thread_index_regex", "(T|t)hread (?P<thread_index>[0-9]+)( |$)")
        )

        def report_at_verbosity(key, lvl):
            if self.verbose >= lvl and key not in self.report:
                self.report.append(key)

        self.verbose = self.verbose if "verbose" not in kwargs else kwargs["verbose"]
        self.report = kwargs["report"] if "report" in kwargs else self.report
        report_at_verbosity("category", 1)
        report_at_verbosity("process", 2)
        report_at_verbosity("threads", 2)
        report_at_verbosity("track_ids", 2)

        _filenames = sorted(self.filename)
        self.filename = kwargs.get("filename", sorted(self.filename))

        _new_filenames = [x for x in self.filename if x not in _filenames]

        if len(self.filename) + len(_new_filenames) != len(self.trace_processor):
            with self.timer.phase("TP construction"):
                self.trace_processor = [
                    TraceProcessor(trace=(f)) for f in self.filename
                ]
        elif _new_filenames:
            with self.timer.phase("TP construction"):
                self.trace_processor += [
                    TraceProcessor(trace=(f)) for f in _new_filenames
                ]

        self.max_depth = kwargs.get("max_depth", None)

        self.timer.end_phase()

    def query_tp(self, query, index_name=lambda x: "tp_index"):
        """Simplifies querying the trace processor and always adds a
        "tp_index" column for referencing which trace_processor
        generated the data
        """

        def _append_column(df, name, idx):
            """Used to add a tp_index column to our data which is used to identify
            which trace-processor the query results came from"""
            if name and name not in df:
                df.insert(0, name, idx)
            return df

        def _get_dataframe(tp):
            """workaround for bug in TraceProcessor.QueryResultIterator.as_pandas_dataframe()"""
            query_itr = tp.query(f"{query}")
            # the perfetto trace processor query function looks like this:
            #
            #     def query(self, sql: str):
            #         response = self.http.execute_query(sql)
            #         if response.error:
            #             raise TraceProcessorException(response.error)
            #         return TraceProcessor.QueryResultIterator(response.column_names,
            #                                                   response.batch)
            #
            # unfortunately, data type of response.column_names is RepeatedScalarContainer
            # and in a lot of versions of pandas, this type does not satisfy any of it's
            # checks for whether this is a valid Index-type for the columns:
            #   isinstance(..., Index)
            #   isinstance(..., ABCSeries)
            #   is_iterator(...)
            #   isinstance(..., list)
            #
            # and thus as_pandas_dataframe() raises a TypeError exception. Queries can
            # be VERY expensive for a large database (>> 10s of seconds) so instead of
            # try -> except -> re-query (necessary) -> convert to list -> re-call
            # as_pandas_dataframe() like so:
            #
            #   try:
            #       return query_itr.as_pandas_dataframe()
            #   except TypeError:
            #       query_itr = tp.query(...)
            #       query_itr.__dict__["..."] = list(query_itr.__dict__["..."])
            #       return query_itr.as_pandas_dataframe()
            #
            # which would effectively result in two queries every single time, we just
            # do it upfront if the dict entry that is known to cause the problem exists
            # and is of the type that we know causes problems
            #
            _buggy_dict_entry = "_QueryResultIterator__column_names"
            if (
                _buggy_dict_entry in query_itr.__dict__
                and type(query_itr.__dict__[_buggy_dict_entry]).__name__
                == "RepeatedScalarContainer"
            ):
                query_itr.__dict__[_buggy_dict_entry] = list(
                    query_itr.__dict__[_buggy_dict_entry]
                )

            return query_itr.as_pandas_dataframe()

        with self.timer.phase("query tp"):
            return pd.concat(
                [
                    _append_column(
                        _get_dataframe(tp),
                        index_name(idx),
                        idx,
                    )
                    for idx, tp in enumerate(self.trace_processor)
                ]
            )

    def extract_tp_data(self, **kwargs):
        """Extracts all the necessary data from the trace processor"""
        self.configure(**kwargs)

        self.timer.start_phase("PerfettoReader query")

        with self.timer.phase("TP slice query"):
            self.dataframe = self.query_tp(
                "SELECT slice_id, track_id, category, depth, stack_id, parent_stack_id, ts, dur, name FROM slice"
            )

        with self.timer.phase("category filter gen"):
            self.df_categories = sorted(list(self.dataframe["category"].unique()))

            # check for update to include/exclude category
            self.exclude = kwargs.get("exclude_category", self.exclude)
            self.include = kwargs.get("include_category", self.include)

            self.categories = self.df_categories[:]

            # apply include first
            if self.include:
                self.categories = [x for x in self.categories if x in self.include]

            # apply exclude after
            if self.exclude:
                self.categories = [x for x in self.categories if x not in self.exclude]

            self.default_categories = kwargs.get(
                "default_categories", self.default_categories
            )
            _acceptable_default_categories = 'default_categories can be set to: "all", ["all"], or [list of categories...]'

            if not self.categories and self.default_categories:
                if not isinstance(self.default_categories, (tuple, list)):
                    self.default_categories = [self.default_categories]

                if "all" in self.default_categories:
                    self.categories = self.df_categories[:]
                else:
                    raise ValueError(
                        f"invalid default_categories value: {self.default_categories}. {_acceptable_default_categories}"
                    )

            # filter out any categories that do not exist
            self.categories = sorted(
                [x for x in self.categories if x in self.df_categories]
            )

            if not self.categories:
                raise ValueError(
                    f"The application of include_category={self.include} followed by exclude_category={self.exclude} rendered an empty set of categories (available={self.df_categories}). Either clear one of the configs or assign the default_categories. {_acceptable_default_categories}"
                )

            if "category" in self.report:
                _ignore = [x for x in self.df_categories if x not in self.categories]
                print(
                    "categories: {}{}".format(
                        ", ".join(self.categories),
                        " (ignored: {})".format(", ".join(_ignore)) if _ignore else "",
                    )
                )

            # reduce the dataframe to given specified category data
            # TODO: adjust the parent stack ids. if <user> category entry is child of <host> category entry, we lose <user> category entry
            self.dataframe = self.dataframe[
                self.dataframe["category"].isin(self.categories)
            ]

            if self.dataframe.empty:
                raise RuntimeError(
                    "category filtering resulted in an empty dataframe. categories: include={}, exclude={}, available={}".format(
                        self.include, self.exclude, self.df_categories
                    )
                )

        with self.timer.phase("TP metadata query"):
            self.process = self.query_tp(
                "SELECT process.upid AS process_upid, process.id AS process_id, process.pid, process.name AS process_name, process_track.upid as track_upid, process_track.id AS track_id, process_track.parent_id as track_parent_id, process_track.name AS track_name from process JOIN process_track ON process_track.upid = process.upid WHERE process.pid > 0"
            )
            self.threads = self.query_tp(
                "SELECT thread.utid AS thread_utid, thread.id AS thread_id, thread.tid, thread.name as thread_name, thread.is_main_thread, thread_track.id AS track_id, thread_track.parent_id AS track_parent_id, thread_track.name AS track_name from thread JOIN thread_track ON thread_track.utid = thread.utid"
            )

        with self.timer.phase("track ids gen"):
            # generate empty dictionaries for each trace processor
            self.track_ids = [{} for _ in range(len(self.trace_processor))]

            # generate mapping from track IDs to process and thread info.
            # the "pid" and "tid" fields are the system value. we want to
            # assign a "rank" and "thread" value for "pid" and "tid",
            # respectively which start at zero and monotonically increase
            for thread in self.threads.itertuples():
                _thread_name = (
                    thread.thread_name
                    if thread.track_name is None
                    else thread.track_name
                )
                for process in self.process.itertuples():
                    if process.tp_index != thread.tp_index:
                        continue
                    _process_name = (
                        process.process_name
                        if process.track_name is None
                        else process.track_name
                    )
                    if process.track_id == thread.track_parent_id:
                        self.track_ids[thread.tp_index][thread.track_id] = {
                            "tp_index": thread.tp_index,
                            "pid": process.pid,
                            "tid": thread.tid,
                            "rank": -1,
                            "thread": -1,
                            "prio": 0 if thread.is_main_thread else 1,
                            "process_name": _process_name,
                            "thread_name": _thread_name,
                        }
                        break

            # some track ids do not have an associated system thread id so handle them here.
            # for example, omnitrace post-processes sampling data collected on a thread
            # during finalization and is inserted into perfetto on the main thread
            # but not in the main thread track so perfetto does not associate the
            # track_id with a system thread id.
            for process in self.process.itertuples():
                if process.track_id in self.track_ids[process.tp_index].keys():
                    continue
                _process_name = (
                    process.track_name
                    if process.process_name is None
                    else process.process_name
                )
                _thread_name = (
                    process.process_name
                    if process.track_name is None
                    else process.track_name
                )
                self.track_ids[process.tp_index][process.track_id] = {
                    "tp_index": process.tp_index,
                    "pid": process.pid,
                    "tid": process.pid,
                    "rank": -1,
                    "thread": -1,
                    "prio": 0 if process.track_parent_id is None else 2,
                    "process_name": _process_name,
                    "thread_name": _thread_name,
                }

        if "track_ids" in self.report and self.verbose >= 3:
            print("\ntrack ids (original):")
            for idx, _track_ids in enumerate(self.track_ids):
                for key, itr in _track_ids.items():
                    print(f"  {idx:2}:  {key:8} :: {itr}")
            print("")

        with self.timer.phase("PID/TID index gen"):
            # since the protobuf just has raw (system) PID and TIDs and there may be multiple PIDs and TIDs
            # in the same file, we need to map the system PIDs to rank IDs starting at zero and, for each
            # PID, map the system TIDs to thread-ids starting at zero
            pid_offset = 0
            for idx, _track_ids in enumerate(self.track_ids):
                _track_ids = dict(
                    sorted(
                        _track_ids.items(),
                        key=lambda x: [x[1]["pid"], x[1]["prio"], x[1]["tid"]],
                    )
                )
                pids = list(set([x["pid"] for _, x in _track_ids.items()]))

                if self.verbose >= 3:
                    tids = list(set([x["tid"] for _, x in _track_ids.items()]))
                    print(f"pids: {pids}")
                    print(f"tids: {tids}")

                # assign the rank and then increment the rank offset by the number of PIDs in the file
                for pidx, pid in enumerate(pids):
                    for _, itr in _track_ids.items():
                        if itr["pid"] == pid:
                            itr["rank"] = pidx + pid_offset
                pid_offset += len(pids)

                for _, pid in enumerate(pids):
                    # dictionary containing only the data for this pid
                    _pid_track_ids = dict(
                        [[x, y] for x, y in _track_ids.items() if y["pid"] == pid]
                    )

                    assigned_track_ids = []

                    # filter out the main threads (priority == 0) for a given pid and set index to a value of zero
                    main_thr_info = set(
                        [x for x, y in _pid_track_ids.items() if y["prio"] == 0]
                    )

                    # for known "main" threads, assign index to zero
                    for track_id, track_id_data in _track_ids.items():
                        if track_id in main_thr_info:
                            track_id_data["thread"] = 0
                            assigned_track_ids.append(track_id)

                    # starting value for assignment. set before next step
                    offset = 1 if assigned_track_ids else 0

                    # search thread name to try to identify which thread it belongs to.
                    # needs to come after offset assignment
                    for track_id, track_id_data in _track_ids.items():
                        if (
                            track_id in assigned_track_ids
                            or track_id not in _pid_track_ids.keys()
                        ):
                            continue
                        m = re.search(
                            self.thread_index_regex, track_id_data["thread_name"]
                        )
                        if m:
                            track_id_data["thread"] = int(m.group("thread_index"))
                            assigned_track_ids.append(track_id)

                    # filter out the non-main threads (priority > 0) for a given pid that haven't already been assigned an index
                    chld_thr_info = set(
                        [
                            x
                            for x, y in _pid_track_ids.items()
                            if y["prio"] > 0 and x not in assigned_track_ids
                        ]
                    )

                    # finally, assign remaining tracks thread indexes via incrementing offset value
                    for track_id, track_id_data in _track_ids.items():
                        if (
                            track_id in assigned_track_ids
                            or track_id not in _pid_track_ids.keys()
                        ):
                            continue
                        if track_id in chld_thr_info:
                            track_id_data["thread"] = offset
                            assigned_track_ids.append(track_id)
                            offset += 1

                    # make sure the thread indexes are monotonically increasing
                    # this may not be the case because of the assignment via regex matching
                    _pid_track_ids = dict(
                        [[x, y] for x, y in _track_ids.items() if y["pid"] == pid]
                    )
                    tidx_max = max([y["thread"] for x, y in _pid_track_ids.items()])
                    tidx_uniq = len(
                        set(
                            [
                                y["thread"]
                                for x, y in _pid_track_ids.items()
                                if y["thread"] >= 0
                            ]
                        )
                    )

                    if self.verbose >= 3:
                        print(f"\nTID :: max={tidx_max}, unique={tidx_uniq}\n")

                    # add one to comparison since one thread with a value of 0 would be a size of 1
                    while tidx_max + 1 > tidx_uniq:
                        for idx in range(tidx_max):
                            # if this is empty, we need to decrement all thread indexes > idx
                            _tidx_loc = [
                                x
                                for x, y in _pid_track_ids.items()
                                if y["thread"] == idx
                            ]
                            if not _tidx_loc:
                                for itr in _pid_track_ids.keys():
                                    if _track_ids[itr]["thread"] > idx:
                                        _track_ids[itr]["thread"] -= 1
                                break
                                # exit the loop so that we recalculate tidx_max
                        tidx_max = max([y["thread"] for x, y in _pid_track_ids.items()])

        if "process" in self.report:
            print(f"\nprocess:\n{self.process.to_string()}\n")
        if "threads" in self.report:
            print(f"\nthreads:\n{self.threads.to_string()}\n")
        if "track_ids" in self.report:
            print("\ntrack ids:")
            for idx, _track_ids in enumerate(self.track_ids):
                for key, itr in _track_ids.items():
                    print(f"  {idx:2}:  {key:8} :: {itr}")
            print("")

        if self.verbose >= 3:
            print("\nTID mapping:")
            for idx, _track_ids in enumerate(self.track_ids):
                for track_id, itr in _track_ids.items():
                    pid = itr["pid"]
                    tid = itr["tid"]
                    pidx = itr["rank"]
                    tidx = itr["thread"]
                    print(
                        f"  {idx:2}:  [{track_id:4}] {pid:8} -> {pidx:8} :: {tid:8} -> {tidx:8}"
                    )

                print("")

        self.timer.end_phase()

    def create_graph(self):
        """Create graph and dataframe"""

        @self.timer.decorator("frame attributes")
        def get_frame_attributes(_name):
            """Get the standard set of dictionary entries for a Frame.
            Also, parses the prefix for func-file-line info
            which is typically in the form:
                <FUNC> [<FILE>:<LINE>]
                <FUNC> [<FILE>]
            """

            if not self.compiled_patterns:
                return {"type": "function", "name": _name}

            def _process_regex(_data):
                """Process the regex data for func/file/line info"""
                return _data.groupdict() if _data is not None else None

            def _perform_regex(_name):
                """Performs a search for standard configurations of function + file + line"""
                for _pattern in self.compiled_patterns:
                    _tmp = _process_regex(re.search(_pattern, _name))
                    if _tmp:
                        return _tmp
                return None

            _keys = {"type": "region", "name": _name}
            _extra = {"file": "<unknown>", "line": "0"}
            _pdict = _perform_regex(_name)
            if _pdict is not None:
                _func = _pdict.get("func", None)
                _file = _pdict.get("file", "<unknown>")
                _line = _pdict.get("line", "0")
                _head = _pdict.get("head", None)
                _tail = _pdict.get("tail", None)

                _line_s = f":{_line}" if int(_line) > 0 else ""
                _tail_s = f"/{_tail}" if _tail is not None else ""
                _file_s = f"{_file}{_line_s}" if _file != "<unknown>" else _file

                _extra["file"] = _file_s
                _extra["line"] = _line

                if "head" in _pdict:
                    _keys["name"] = _head.rstrip()
                    if _func is not None:
                        _extra["func"] = _func
                else:
                    if _func is not None:
                        _keys["name"] = _func
                    else:
                        _keys["name"] = _file_s

                _keys["name"] = "{}{}".format(_keys["name"], _tail_s)

            return (_keys, _extra)

        list_roots = []
        track_id_dict = OrderedDict()
        callpath_to_node = {}

        def_metric = self.default_metric

        df = self.dataframe
        _cols = [
            "tp_index",
            "track_id",
            "category",
            "slice_id",
            "stack_id",
            "parent_stack_id",
            "name",
            "depth",
            "ts",
            "dur",
        ]
        _data = [df[x].to_list() for x in _cols]

        assert min([len(x) for x in _data]) == max([len(x) for x in _data])

        self.timer.start_phase("slice processing")

        for _tp_index, _track_id in zip(
            _data[_cols.index("tp_index")], _data[_cols.index("track_id")]
        ):
            assert _tp_index < len(self.track_ids)
            assert _track_id in self.track_ids[_tp_index].keys()

            _track_info = self.track_ids[_tp_index][_track_id]
            _rank = _track_info["rank"]
            _thread = _track_info["thread"]

            if _tp_index not in track_id_dict:
                track_id_dict[_tp_index] = OrderedDict()
            if _rank not in track_id_dict[_tp_index]:
                track_id_dict[_tp_index][_rank] = OrderedDict()
            if _thread not in track_id_dict[_tp_index][_rank]:
                track_id_dict[_tp_index][_rank][_thread] = OrderedDict()

            track_id_dict[_tp_index][_rank][_thread] = {0: None}

        for (
            _tp_index,
            _track_id,
            _category,
            _slice_id,
            _stack_id,
            _parent_stack_id,
            _name,
            _depth,
            _ts,
            _dur,
        ) in zip(*_data):
            _track_info = self.track_ids[_tp_index][_track_id]
            _rank = _track_info["rank"]
            _thread = _track_info["thread"]

            _track_id_dict = track_id_dict[_tp_index][_rank][_thread]

            # removed because of filtering
            if _parent_stack_id not in _track_id_dict:
                continue

            # reduce processing time
            if self.max_depth is not None and _depth > self.max_depth:
                continue

            _metrics = {}
            _metrics["rank"] = _track_info["rank"]
            _metrics["thread"] = _track_info["thread"]
            _metrics["pid"] = _track_info["pid"]
            _metrics["tid"] = _track_info["tid"]
            _metrics["track_id"] = _track_id
            _metrics["slice_id"] = _slice_id
            _metrics["stack_id"] = _stack_id
            _metrics["parent_stack_id"] = _parent_stack_id
            _metrics["ts"] = _ts
            _metrics[def_metric] = float(_dur) * 1.0e-9  # nsec -> sec

            _frame_attrs, _extra = get_frame_attributes(_name)
            _extra["tp_index"] = _tp_index
            _extra["category"] = _category
            _extra["depth"] = _depth

            # look up the parent node specific to the TP index, rank, and thread
            # stack ID is assigned by perfetto and parent stack ID is the
            # stack ID of it's parent.
            _parent_node = _track_id_dict[_parent_stack_id]

            hnode = Node(Frame(_frame_attrs, **_extra), None)

            if _parent_node:
                _parent_node.add_child(hnode)
            else:
                list_roots.append(hnode)

            # make sure this stack ID is unique for the
            # TP index, rank, and thread and is equal to a
            # previously seen node with the same stack ID
            if _stack_id not in _track_id_dict:
                _track_id_dict[_stack_id] = hnode
            elif _track_id_dict[_stack_id] != hnode:
                _existing = _track_id_dict[_stack_id]
                raise RuntimeError(
                    f"{_stack_id} already exists in track_id_dict[{_tp_index}][{_rank}][{_thread}]. failed to set:\n  {hnode.frame} (current)\n    {_existing.frame} (existing)"
                )

            _hash = hash((_tp_index, _slice_id))
            if _hash in callpath_to_node:
                raise ValueError(f"{_hash} already exists in callpath_to_node dict")

            _frame_attrs.pop("type")  # should not be a column in dataframe
            callpath_to_node[_hash] = dict(
                {"node": hnode, **_frame_attrs},
                **_metrics,
                **_extra,
            )

        self.timer.end_phase()

        with self.timer.phase("graphframe creation"):
            if not callpath_to_node:
                raise RuntimeError(
                    "call-graph is empty. if category filtering was used, you may have filtered out all the root nodes and thus all of it's children"
                )

            graph, dataframe = graphframe_indexing_helper(
                list_roots,
                data=list(callpath_to_node.values()),
                extensions=["rank", "thread"],
                fill_value=0,
            )

            inc_metrics = [self.default_metric]
            exc_metrics = []
            return (graph, dataframe, exc_metrics, inc_metrics, self.default_metric)

    def read(self, **kwargs):
        """Read perfetto json."""

        self.configure(**kwargs)
        self.extract_tp_data(**kwargs)

        with self.timer.phase("graph construction"):
            (
                graph,
                dataframe,
                exc_metrics,
                inc_metrics,
                def_metric,
            ) = self.create_graph()

        if "profile" in self.report:
            print("{}".format(self.timer.to_string("PerfettoReader Performance:\n")))

        def _read(**kwargs):
            self.timer = Timer()
            return self.read(**kwargs)

        def _configure(**kwargs):
            self.configure(**kwargs)

        def _query(*args, **kwargs):
            assert self.trace_processor
            return self.query_tp(*args, **kwargs)

        return GraphFrame(
            graph,
            dataframe,
            exc_metrics,
            inc_metrics,
            default_metric=def_metric,
            metadata=self.metadata,
            attributes={
                "reader": self,
                "read": _read,
                "query": _query,
                "configure": _configure,
                "selected_categories": lambda: self.categories,
                "available_categories": lambda: self.df_categories,
            },
        )



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from collections import OrderedDict
from contextlib import contextmanager
from datetime import datetime
from io import StringIO
from functools import wraps


class Timer(object):
    """Simple phase timer with a context manager."""

    def __init__(self):
        self._phase = None
        self._start_time = None
        self._times = OrderedDict()

    def start_phase(self, phase):
        now = datetime.now()
        delta = None

        if self._phase:
            delta = now - self._start_time
            self._times[self._phase] = delta

        self._phase = phase
        self._start_time = now
        self._times[self._phase] = None
        return delta

    def end_phase(self):
        assert self._phase and self._start_time

        now = datetime.now()
        delta = now - self._start_time
        if self._times.get(self._phase):
            self._times[self._phase] = self._times.get(self._phase) + delta
        else:
            self._times[self._phase] = delta

        self._phase = None
        self._start_time = None

    def to_string(self, header="Times:\n"):
        _times = [[x, y] for x, y in self._times.items() if y is not None]
        _fmt = "    %-{}s %.2fs\n".format(
            min([73, max([20] + [len(x) + 1 for x, y in _times])])
        )
        out = StringIO()
        out.write(header)
        for phase, delta in _times:
            out.write(_fmt % ("{}:".format(phase), delta.total_seconds()))
        return out.getvalue()

    def __str__(self):
        return self.to_string()

    def __iadd__(self, rhs):
        for phase, delta in rhs._times.items():
            if self._times.get(phase):
                self._times[phase] = self._times.get(phase) + delta
            else:
                self._times[phase] = delta
        return self

    @contextmanager
    def phase(self, name):
        _timer = Timer()
        _timer.start_phase(name)
        yield
        _timer.end_phase()
        self += _timer

    def decorator(self, name):
        return TimerDecorator(self, name)


class TimerDecorator(object):
    """Wrapper around Timer to provide a decorator"""

    def __init__(self, timer, name):
        self._timer = timer
        self._name = name

    def __call__(self, func):
        """Decorator"""

        @wraps(func)
        def function_wrapper(*args, **kwargs):
            _timer = Timer()
            _timer.start_phase(self._name)
            result = func(*args, **kwargs)
            _timer.end_phase()
            self._timer += _timer
            return result

        return function_wrapper



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import hatchet as ht
import pstats
import os
from hatchet.util.profiler import Profiler

import pytest
import sys


decorated_calls = [
    "copy",
    "deepcopy",
    "drop_index_levels",
    "filter",
    "squash",
    "tree",
    "to_dot",
    "__add__",
    "__imul__",
    "__isub__",
    "traverse",
    "copy",
    "__len__",
]


def f():
    """Dummy function for profiling"""
    for i in range(1000):
        for j in range(1000):
            i * j


def test_profiler():
    """Test the profiler start/stop, ensure that f() has been profiled"""
    prf = Profiler()

    prf.start()
    f()
    prf.stop()

    sts = pstats.Stats(prf._prf)
    assert "stats" in sts.__dict__

    fn_names = []
    for stat in sts.__dict__["stats"]:
        fn_names.append(stat[2])

    assert "f" in fn_names

    assert isinstance(prf.__str__(), str)
    assert "(f)" in prf.__str__()

    os.remove(prf._output + ".pstats")


def test_write_file():
    """Test that write_to_file, writes a profile to the correct file."""
    prf = Profiler()

    prf.start()
    f()
    prf.stop()

    prf.write_to_file("test.pstats")

    assert os.path.exists("test.pstats")

    sts = pstats.Stats("test.pstats")
    fn_names = []
    for stat in sts.__dict__["stats"]:
        fn_names.append(stat[2])

    assert "f" in fn_names

    os.remove("test.pstats")
    os.remove(prf._output + ".pstats")


@pytest.mark.skipif(sys.version_info > (3, 8), reason="Temporarily allow this to fail.")
def test_profiling_calc_pi(calc_pi_hpct_db):
    """Test debug wrapper as called from hpctoolkit."""
    prf = Profiler()
    output_file = prf._output + ".pstats"

    prf.start()
    gf = ht.GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))

    gf.copy()
    gf2 = gf.deepcopy()
    gf.tree()
    gf.to_dot()

    gf3 = gf + gf2
    gf *= gf2
    gf3 -= gf2

    gf.graph.traverse()
    gf.graph.copy()
    len(gf.graph)

    gf2 = gf.filter(lambda x: x["time"] > 0.01)

    gf2.squash()
    gf.drop_index_levels()

    prf.stop()

    assert os.path.exists(output_file)

    sts = pstats.Stats(output_file)

    fn_names = []
    for stat in sts.__dict__["stats"]:
        fn_names.append(stat[2])

    for call in decorated_calls:
        assert call in fn_names

    os.remove(output_file)



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

# make flake8 unused names in this file.
# flake8: noqa: F401

from .graphframe import GraphFrame
from .query import QueryMatcher
from .chopper import Chopper



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import cython

def subtract_exclusive_metric_vals(long nid, long parent_nid, double[:] metrics, long num_stmt_nodes, long stride):
  cdef long ref_nid = nid
  cdef long ref_pnid = parent_nid

  # compiler directive
  with cython.boundscheck(False):
  # we are modifying metrics in place here
  # since they are passed by refrence via their
  # memory
    for i in range(num_stmt_nodes):
      metrics[ref_pnid-1] -= metrics[ref_nid-1]

      ref_nid += stride
      ref_pnid += stride



# Copyright 2020-2024 The Regents of the University of California, through
# Lawrence Berkeley National Laboratory, and other Hatchet Project Developers.
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np

timemory_avail = True
try:
    from timemory.profiler import Profiler
    from timemory.trace import Tracer
    from timemory.component import WallClock, CpuClock
except ImportError:
    timemory_avail = False

    class FakeComponent:
        def __init__(self, *args, **kwargs):
            pass

        @staticmethod
        def id():
            pass

    class FakeProfiler:
        def __init__(self, *args, **kwargs):
            pass

        def __call__(self, func):
            pass

    WallClock = FakeComponent
    CpuClock = FakeComponent
    Profiler = FakeProfiler
    Tracer = FakeProfiler


ncnt = 0
components = [WallClock.id(), CpuClock.id()]


def eval_func(arr, tol):
    """Dummy tolerance-checking function for profiling which introduces some arbitrary branching"""
    global ncnt
    max = np.max(arr)
    if ncnt % 3 != 0:
        avg = np.mean(arr)
    else:
        avg = max
    ncnt = ncnt + 1
    return True if avg < tol and max < tol else False


@Profiler(components)
def prof_func(arr, tol):
    """Dummy function for profiling"""
    while not eval_func(arr, tol):
        arr = arr - np.power(arr, 3)


@Tracer(components)
def trace_func(arr, tol):
    """Dummy function for tracing"""
    while not eval_func(arr, tol):
        arr = arr - np.power(arr, 3)



# Copyright 2021-2024 University of Maryland and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
import numpy as np


class Chopper:
    """High-level API for performance analysis."""

    def flat_profile(self, graphframe, groupby_column=None, as_index=True):
        """Generates flat profile for a given graphframe.
        Returns a new dataframe."""
        graphframe2 = graphframe.deepcopy()

        if groupby_column is None:
            groupby_column = "name"

        # TODO: change drop_index_levels(). Drop only ranks or threads.
        graphframe2.drop_index_levels()

        result_dataframe = graphframe2.dataframe.groupby(
            groupby_column, as_index=as_index
        ).sum()
        result_dataframe = result_dataframe.sort_values(
            by=[graphframe.default_metric], ascending=False
        )

        return result_dataframe

    def flatten(self, graphframe, groupby_column="name"):
        """
        Flattens the graphframe by changing its graph structure and the dataframe.
        Returns a new graphframe.
        """
        result_graphframe = graphframe.groupby_aggregate(groupby_column, "sum")

        return result_graphframe

    def to_callgraph(self, graphframe):
        """
        Converts a CCT to a callgraph.
        Returns a new graphframe.
        """

        # TODO: provide hierarchy information in the graphframe metadata to access the
        #       hierarchy of the input nodes - currently using function name
        result_graphframe = graphframe.groupby_aggregate("name", "sum")

        return result_graphframe

    def load_imbalance(
        self, graphframe, metric_column=None, threshold=None, verbose=False
    ):
        """Calculates load imbalance for the given metric column.
        Takes a graphframe and a metric column to calculate the
        load imbalance.
        It takes a threshold value to filter out the insignificant nodes
        from the graphframe. The threshold parameter takes a percentage.
        For example, threshold=0.01 on time metric filters out the nodes
        that the program spends less than 1% of the max value of time metric.
        Returns a new graphframe with corresponding <metric>.imbalance column.
        If the verbose parameter is True, it provides frequency histogram,
        the top five ranks that have the highest metric value, and percentile
        information.
        """

        def _update_metric_lists(metric_type, metric):
            """Update graphframe.inc_metrics and graphframe.exc_metrics
            lists after renaming/creating columns"""
            metric_type.append(metric + ".mean")
            metric_type.append(metric + ".max")

        def _calculate_statistics(dataframe, metric_column, func):
            """Calculate frequency histogram and percentiles. Find the
            ranks that have the highest metric values."""
            index_names = list(dataframe.index.names)
            index_names.remove("node")

            # create dict that stores aggregation function for each column.
            # fuc = [np.mean, np.max] for the metric_column.
            agg_dict = {}
            for col in dataframe.columns.tolist():
                if col == metric_column:
                    agg_dict[col] = func
                else:
                    agg_dict[col] = lambda x: x.iloc[0]

            # move node from index to column and group by node.
            dataframe.reset_index(level="node", inplace=True)
            grouped_df = dataframe.groupby("node")

            # if verbose, calculate statistics for the
            # frequency histogram, percentiles, and the top
            # five ranks that has the highest metric value.
            if verbose:
                statistics_dict = {}
                # for each group
                for node, item in grouped_df:
                    freqs = []
                    percentiles = []

                    group = grouped_df.get_group(node)
                    group.reset_index(inplace=True)
                    # sort by metric value. get rank and metric value
                    sorted_rank_metric = group[["rank", metric_column]].sort_values(
                        by=metric_column, ascending=False
                    )
                    # store sorted ranks and metrics
                    sorted_ranks = sorted_rank_metric["rank"].to_numpy()
                    sorted_metric = sorted_rank_metric[metric_column].to_numpy()
                    num_ranks = len(sorted_ranks)

                    # we have fixed number of bins = 10.
                    # Example: max=100, min=0, size=(100-0)/10=10
                    bins = 10
                    max = sorted_metric[0]
                    min = sorted_metric[-1]
                    size = (max - min) / bins
                    if size != 0:
                        for i in range(bins):
                            bin_start = min + i * size
                            bin_end = min + (i + 1) * size
                            # sometimes bin_end != end because of
                            # rounding. For example:
                            # bin_end=52.93999, max = 53.94.
                            if i == bins - 1:
                                bin_end = max
                            if i == 0:
                                bin_start = min

                            # count the number of ranks in a bin
                            count = len(
                                sorted_rank_metric[
                                    (sorted_rank_metric[metric_column] > bin_start)
                                    & (sorted_rank_metric[metric_column] <= bin_end)
                                ]["rank"].unique()
                            )
                            freqs.append(count)

                        # the first rank's value can be equal to min so
                        # we count them as well.
                        freqs[0] += len(
                            sorted_rank_metric[
                                (sorted_rank_metric[metric_column] == min)
                            ]["rank"].unique()
                        )
                    else:
                        # Example: if min=max=0 and num_ranks=2, freqs=[2, 2]
                        freqs = [num_ranks] * bins

                    # calculate percentiles.
                    for i in [0, 25, 50, 75, 100]:
                        percentiles.append(np.percentile(sorted_metric, i))

                    # find the top five ranks that have the
                    # highest metric value.
                    if len(sorted_ranks) > 5:
                        sorted_ranks = sorted_ranks[:5]

                    # create statistics_dict.
                    # add metric_column -> imbalance.ranks, time.hist, time.percentiles
                    statistics_dict[node] = {
                        "node": node,
                        "{}.ranks".format(metric_column): sorted_ranks,
                        "{}.hist".format(metric_column): freqs,
                        "{}.percentiles".format(metric_column): percentiles,
                    }

                # create statistics dataframe from the dict.
                statistics_df = pd.DataFrame.from_dict(statistics_dict.values())
                statistics_df.set_index("node", inplace=True)

            # aggregate grouped dataframe
            agg_df = grouped_df.agg(agg_dict)

            # pandas creates multiindex columns when we do .agg() using
            # multiple functions (np.mean and np.max).
            # We first flatten the columns to remove multiindex.
            # Example after to_flat_index():
            # [('time', 'mean'), ('time', 'max'), ('name', '<lambda>')]
            # Then remove rename metric_column by adding mean and max.
            # Remove <lambda> from others.
            agg_df.columns = agg_df.columns.to_flat_index()
            columns = agg_df.columns.values

            for idx in range(len(columns)):
                if columns[idx][0] == metric_column and columns[idx][1] == "mean":
                    columns[idx] = metric_column + ".mean"
                elif columns[idx][0] == metric_column and columns[idx][1] == "max":
                    columns[idx] = metric_column + ".max"
                else:
                    columns[idx] = columns[idx][0]

            # if verbose, join aggregate dataframe with
            # statistics dataframe.
            if verbose:
                agg_df = agg_df.join(statistics_df, how="outer")

            return agg_df

        # Create a copy of the GraphFrame.
        graphframe2 = graphframe.deepcopy()

        # Use default_metric if not given.
        if metric_column is None:
            metric_column = graphframe.default_metric

        assert (
            metric_column in graphframe2.dataframe.columns
        ), "{} column does not exist in the dataframe.".format(metric_column)

        if threshold is not None:
            # Get the max value of the given metric.
            max_val = graphframe2.dataframe.sort_values(by=metric_column).iloc[-1][
                metric_column
            ]
            # Calculate the threshold.
            thres_val = max_val * threshold

        # Similar to drop_index_levels() but it calculates
        # statistics if verbose == True.
        graphframe2.dataframe = _calculate_statistics(
            graphframe2.dataframe, metric_column, ["mean", "max"]
        )

        graphframe2.inc_metrics = []
        graphframe2.exc_metrics = []

        # Add new columns to .inc_metrics or .exc_metrics.
        if metric_column in graphframe2.inc_metrics:
            _update_metric_lists(graphframe2.inc_metrics, metric_column)
        elif metric_column in graphframe2.exc_metrics:
            _update_metric_lists(graphframe2.exc_metrics, metric_column)

        # filter out the nodes if their max metric value across
        # processes/threads is less than threshold% of the max value
        # of the given metric.
        if threshold is not None:
            graphframe2 = graphframe2.filter(
                lambda x: x[metric_column + ".max"] > thres_val
            )

        # Calculate load imbalance for the given metric
        # by calculating max-to-mean ratio.
        graphframe2.dataframe[metric_column + ".imbalance"] = graphframe2.dataframe[
            metric_column + ".max"
        ].div(graphframe2.dataframe[metric_column + ".mean"])

        # <metric_column>.max is already stored in <metric_column>.percentiles
        # column if verbose = True. We don't drop it if verbose = False.
        if verbose:
            graphframe2.dataframe.drop(metric_column + ".max", axis=1, inplace=True)

        # default metric will be imbalance when user print the tree
        graphframe2.default_metric = metric_column + ".imbalance"
        # sort by mean value
        graphframe2.dataframe = graphframe2.dataframe.sort_values(
            by=[metric_column + ".mean"], ascending=False
        )
        return graphframe2

    def hot_path(
        self, graphframe, start_node=None, metric=None, threshold=0.5, callpath=[]
    ):
        """Returns the hot_path function.
        Inputs:
         - start_node (optional): Start node of the hot path can be given.
         Default: the root node that has the largest metric value.
         - metric: A numerical metric on the dataframe.
         Default: graphframe.default_metric
         - threshold: Threshold for parent-child comparison (parent <= child/2).
         Default: 0.5
        Output:
         - hot_path: list of nodes, starting from the start node to the hot node.

        Example:
        root_node = graphframe.graph.roots[0]
        graphframe.hot_path(root_node)
        """

        def find_hot_path(graphframe, parent, metric, threshold, callpath):
            parent_metric = graphframe.dataframe.loc[parent, metric]
            sorted_child_metric = []
            # Get all children nodes with their metric values and append
            # them to a list.
            for child in parent.children:
                child_metric = graphframe.dataframe.loc[child, metric]
                sorted_child_metric.append((child, child_metric))

            if sorted_child_metric:
                # in-place sort children by their metric values.
                sorted_child_metric.sort(key=lambda x: x[1], reverse=True)
                child = sorted_child_metric[0][0]
                child_metric = sorted_child_metric[0][1]
                if child_metric < (threshold * parent_metric):
                    # return hotpath since child's metric value is
                    # not greater than threshold * parent's metric.
                    callpath.append(child)
                    return callpath
                else:
                    # continue from child if its metric is greater than
                    # threshold * parent's metric.
                    # For example, child_metric >= parent_metric/2
                    callpath.append(child)
                    return find_hot_path(graphframe, child, metric, threshold, callpath)

            return callpath

        # copy the graphframe not to modify the original graph
        gf_copy = graphframe.deepcopy()
        gf_copy.drop_index_levels()

        # choose the default metric if metric has not set
        if metric is None:
            metric = graphframe.default_metric

        # choose the root node that has the greatest metric value
        # if a start node is not specified
        if start_node is None:
            roots_metrics = []
            for root in gf_copy.graph.roots:
                roots_metrics.append((root, gf_copy.dataframe.loc[root, metric]))

            # in-place sort
            roots_metrics.sort(key=lambda x: x[1], reverse=True)
            start_node = roots_metrics[0][0]

        return find_hot_path(
            gf_copy, start_node, metric, threshold, callpath=[start_node]
        )

    @staticmethod
    def multirun_analysis(
        graphframes=[],
        pivot_index="num_processes",
        columns=["name"],
        metric="time",
        threshold=None,
        groupby_function=None,
    ):
        """Creates a pivot table.
        Inputs:
         - graphframes: A list of graphframes.
         - pivot_index: The metric in each graphframe's metadata used to index the pivot table.
         Default: num_processes
         - columns: The non-numerical metric over which the pivot table's column values are aggregated.
         Default: name
         - metric: The numerical metric which is aggregated to form the pivot table's column values.
         Default: time
         - threshold: The threshold for filtering metric rows of the graphframes.
        Output:
         - a pivot table
        """

        assert (
            graphframes is not None and len(graphframes) >= 2
        ), "function param 'graphframes' requires at least two graphframe objects"

        if not isinstance(columns, list):
            columns = [columns]

        for gf in graphframes:
            assert (
                pivot_index in gf.metadata.keys()
            ), "{} missing from GraphFrame metadata: use update_metadata() to specify.".format(
                pivot_index
            )
            assert (
                metric in gf.dataframe.columns
            ), "{} metric not present in all graphframes".format(metric)
            for column in columns:
                assert (
                    column in gf.dataframe.columns
                ), "{} column not present in all graphframes".format(column)

        dataframes = []
        for gf in graphframes:
            gf_copy = gf.deepcopy()
            gf_copy.drop_index_levels()

            # group by name if the user gives a function such as np.mean
            if groupby_function is not None:
                gf_copy.dataframe = gf_copy.dataframe.groupby(
                    "name", as_index=False
                ).agg(groupby_function)

            # Grab the pivot_index from the metadata, store this as a new
            # column in the DataFrame.
            pivot_val = gf.metadata[pivot_index]
            gf_copy.dataframe[pivot_index] = pivot_val

            # Filter the dataframe, keeping only the rows that are above the threshold
            if threshold is not None:
                filtered_rows = gf_copy.dataframe.apply(
                    lambda x: x[metric] > threshold, axis=1
                )
                gf_copy.dataframe = gf_copy.dataframe[filtered_rows]

            # Insert the graphframe's dataframe into a list.
            dataframes.append(gf_copy.dataframe)

        # Concatenate all DataFrames into a single DataFrame called result.
        result = pd.concat(dataframes)

        pivot_df = result.pivot_table(index=pivot_index, columns=columns, values=metric)

        return pivot_df

    @staticmethod
    def speedup_efficiency(
        graphframes=[],
        weak=False,
        strong=False,
        efficiency=False,
        speedup=False,
        pivot_index="num_processes",
        metrics=["time"],
        threshold=None,
    ):
        """
        Calculates the speedup and efficiency values.
        Inputs:
         - graphframes: A list of graphframes.
         - weak: True for weak scaling experiments.
         - strong: True for strong scaling experiments.
         - efficiency: True if the user wants to calculate efficiency.
         - strong: True if the user wants to calculate speedup.
         - pivot_index: The metric in each graphframe's metadata used to do calculations.
         Default: num_processes.
         - metric: The numerical metric for which we want to calculate speedup and efficiency.
         Default: time
         - threshold: The threshold for filtering metric rows of the graphframes.
        Output:
         - a new dataframe that stores speedup and efficiency values.
        """
        from .graphframe import GraphFrame

        assert (
            strong is True or weak is True
        ), "at least one of the 'strong' and 'weak' parameters should be True."
        assert (
            efficiency is True or speedup is True
        ), "at least one of the 'efficiency' and 'speedup' parameters should be True."
        assert (
            weak is False or speedup is False
        ), "speed up can be calculated only for strong scaling."

        process_to_gf = []
        for gf in graphframes:
            assert (
                "num_processes" in gf.metadata.keys()
            ), "pivot_index missing from GraphFrame metadata: use update_metadata() to specify."
            process_to_gf.append((gf.metadata[pivot_index], gf))

        GraphFrame.unify_multiple_graphframes(graphframes)

        sorted(process_to_gf, key=lambda x: x[0])
        base_numpes = process_to_gf[0][0]
        base_graphframe = process_to_gf[0][1]

        result_df = pd.DataFrame()
        # add base values to the resulting dataframe.
        for column in base_graphframe.dataframe.columns:
            if column not in base_graphframe.inc_metrics + base_graphframe.exc_metrics:
                result_df[column] = base_graphframe.dataframe[column]

        # calculate speedup and efficiency.
        for other in process_to_gf[1:]:
            for metric in metrics:
                if weak:
                    new_column_name = "{}.{}.{}".format(other[0], metric, "efficiency")
                    # weak scaling efficiency: base / other
                    result_df[new_column_name] = (
                        base_graphframe.dataframe[metric] / other[1].dataframe[metric]
                    )
                else:
                    if speedup:
                        new_column_name = "{}.{}.{}".format(other[0], metric, "speedup")
                        # strong scaling speedup: base / other
                        result_df[new_column_name] = (
                            base_graphframe.dataframe[metric]
                            / other[1].dataframe[metric]
                        )
                    if efficiency:
                        new_column_name = "{}.{}.{}".format(
                            other[0], metric, "efficiency"
                        )
                        # strong scaling efficiency: base * num_procs_base / other
                        result_df[new_column_name] = (
                            base_graphframe.dataframe[metric] * base_numpes
                        ) / (other[1].dataframe[metric] * other[0])
        return result_df

    def correlation_analysis(self, graphframe, metrics=None, method="spearman"):
        """
        Calculates correlation between metrics of a given graphframe.
        Returns the correlation matrix.
        Pandas provides three different methods: pearson, spearman, kendall
        """
        if not isinstance(metrics, list):
            metrics = [metrics]

        assert len(metrics) > 1, "This function requires at least two metrics."

        for metric in metrics:
            assert (
                metric in graphframe.dataframe.columns
            ), "{} column not present in graphframe".format(metric)

        dataframe = graphframe.dataframe[metrics]
        corr_matrix = dataframe.corr(method=method)
        return corr_matrix



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import re

import pandas as pd
import pydot

import hatchet.graphframe
from ..node import Node
from ..graph import Graph
from ..frame import Frame
from ..util.timer import Timer
from ..util.config import dot_keywords


class GprofDotReader:
    """Read in gprof/callgrind output in dot format generated by gprof2dot."""

    def __init__(self, filename):
        self.dotfile = filename

        self.name_to_hnode = {}
        self.name_to_dict = {}

        self.timer = Timer()

    def create_graph(self):
        """Read the DOT files to create a graph."""
        graphs = pydot.graph_from_dot_file(self.dotfile, encoding="utf-8")

        for graph in graphs:
            for edge in graph.get_edges():
                src_name = edge.get_source().strip('"')
                dst_name = edge.get_destination().strip('"')

                src_hnode = self.name_to_hnode.get(src_name)
                if not src_hnode:
                    # create a node if it doesn't exist yet
                    src_hnode = Node(
                        Frame({"type": "function", "name": src_name}), None
                    )
                    self.name_to_hnode[src_name] = src_hnode

                dst_hnode = self.name_to_hnode.get(dst_name)
                if not dst_hnode:
                    # create a node if it doesn't exist yet
                    dst_hnode = Node(
                        Frame({"type": "function", "name": dst_name}), src_hnode
                    )
                    self.name_to_hnode[dst_name] = dst_hnode
                # add source node as parent
                dst_hnode.add_parent(src_hnode)

                # add destination node as child
                src_hnode.add_child(dst_hnode)

            for node in graph.get_nodes():
                node_name = node.get_name()

                if node_name not in dot_keywords:
                    node_name = node_name.strip('"')
                    hnode = self.name_to_hnode.get(node_name)
                    if not hnode:
                        # create a node if it doesn't exist yet
                        hnode = Node(
                            Frame({"type": "function", "name": node_name}), None
                        )
                        self.name_to_hnode[node_name] = hnode

                    if node.obj_dict["attributes"].get("label") is None:
                        continue
                    node_label = node.obj_dict["attributes"].get("label").strip('"')

                    label_parts = node_label.split(r"\n")
                    if len(label_parts) == 5:
                        module, _, inc, exc, _ = label_parts
                    elif len(label_parts) == 4:
                        _, inc, exc, _ = label_parts
                        module = None
                    else:
                        raise ValueError("invalid node label in dot file")

                    inc_time = float(re.sub(r"\%", "", inc))
                    exc_time = float(re.sub(r"[\(\%\)]", "", exc))

                    # create a dict with node properties
                    node_dict = {
                        "module": module,
                        "name": node_name,
                        "time (inc)": inc_time,
                        "time": exc_time,
                        "node": hnode,
                    }
                    self.name_to_dict[node_name] = node_dict

        # add all nodes with no parents to the list of roots
        list_roots = []
        for key, val in self.name_to_hnode.items():
            if not val.parents:
                list_roots.append(val)

        return list_roots

    def read(self):
        """Read the DOT file generated by gprof2dot to create a graphframe. The DOT file
        contains a call graph.
        """
        with self.timer.phase("graph construction"):
            roots = self.create_graph()
            graph = Graph(roots)
            graph.enumerate_traverse()

        with self.timer.phase("data frame"):
            dataframe = pd.DataFrame.from_dict(data=list(self.name_to_dict.values()))
            index = ["node"]
            dataframe.set_index(index, inplace=True)
            dataframe.sort_index(inplace=True)

        return hatchet.graphframe.GraphFrame(graph, dataframe, ["time"], ["time (inc)"])



# Copyright 2022-2024 The University of Arizona and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import os
import json
import glob
import pytest

from hatchet import GraphFrame
from hatchet.util.logger import Logger


def test_no_logging_by_default(calc_pi_hpct_db, lulesh_caliper_json):
    # test HPCTOOLKIT
    GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))
    assert Logger._active is False

    # test Caliper
    GraphFrame.from_caliper(str(lulesh_caliper_json))
    assert Logger._active is False


@pytest.mark.xfail(reason="Allow this to fail until global configuration PR is merged.")
def test_output(calc_pi_hpct_db):
    functions = [
        "from_hpctoolkit",
        "copy",
        "deepcopy",
        "filter",
        "squash",
        "to_dot",
        "to_flamegraph",
        "to_literal",
        "add",
        "sub",
        "mul",
    ]
    log_lines = []

    gf = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))
    assert Logger._active is True

    gf2 = gf.copy()
    gf3 = gf2.deepcopy()
    filtered = gf3.filter(lambda x: x["time"] > 0.1, squash=False)
    filtered.squash()
    gf.to_dot()
    gf.to_flamegraph()
    gf.to_literal()

    gf3 = gf + gf2
    gf3 = gf - gf2
    gf3 = gf * gf2

    path = "hatchet_*.log"
    files = glob.glob(path)
    assert len(files) == 1

    with open(files[0], "r") as f:
        line = f.readline()
        while line:
            log_lines.append(json.loads(line))
            line = f.readline()

    for line in log_lines:
        assert line["function"] in functions

    Logger.set_inactive()
    os.remove(files[0])



# Copyright 2020-2024 The Regents of the University of California, through
# Lawrence Berkeley National Laboratory, and other Hatchet Project Developers.
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np

from hatchet import GraphFrame
from hatchet.external.console import ConsoleRenderer

import pytest

timemory_avail = True
try:
    import timemory
except ImportError:
    timemory_avail = False


@pytest.mark.skipif(not timemory_avail, reason="timemory package not available")
def test_graphframe(timemory_json_data):
    """Sanity test a GraphFrame object with known data."""
    from timemory.component import WallClock

    wc_s = WallClock.id()  # string identifier
    wc_v = WallClock.index()  # enumeration id
    gf = GraphFrame.from_timemory(timemory_json_data, [wc_s])

    assert len(gf.dataframe) == timemory.size([wc_v])[wc_v]

    for col in gf.dataframe.columns:
        if col in ("sum.wall_clock.inc", "sum.wall_clock"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("nid", "rank"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "node"):
            assert gf.dataframe[col].dtype == object


@pytest.mark.skipif(not timemory_avail, reason="timemory package not available")
def test_tree(timemory_json_data):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_timemory(timemory_json_data)

    print(gf.tree("sum.wall_clock"))

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="sum.wall_clock",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )

    print(output)

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="sum.wall_clock",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )

    print(output)


@pytest.mark.skipif(not timemory_avail, reason="timemory package not available")
def test_graphframe_to_literal(timemory_json_data):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_timemory(timemory_json_data)
    graph_literal = gf.to_literal()

    assert len(graph_literal) == len(gf.graph.roots)


@pytest.mark.skipif(not timemory_avail, reason="timemory package not available")
def test_default_metric(timemory_json_data):
    """Validation test for GraphFrame object using default metric field"""
    gf = GraphFrame.from_timemory(timemory_json_data)

    for func in ["tree", "to_dot", "to_flamegraph"]:
        lhs = "{}".format(getattr(gf, func)(gf.default_metric))
        rhs = "{}".format(getattr(gf, func)())
        assert lhs == rhs
